# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import logging
from logging import Logger
from unittest.mock import patch

import pandas as pd
from ax.core.base_trial import TrialStatus
from ax.core.data import Data
from ax.modelbridge.generation_strategy import (
    GenerationNode,
    GenerationStep,
    GenerationStrategy,
)
from ax.modelbridge.model_spec import ModelSpec
from ax.modelbridge.registry import Models
from ax.modelbridge.transition_criterion import (
    AutoTransitionAfterGenCriterion,
    MaxGenerationParallelism,
    MaxTrials,
    MinimumPreferenceOccurances,
    MinimumTrialsInStatus,
    MinTrials,
    TransitionCriterion,
)
from ax.utils.common.logger import get_logger
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment, get_experiment

logger: Logger = get_logger(__name__)


class TestTransitionCriterion(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.sobol_model_spec = ModelSpec(
            model_enum=Models.SOBOL,
            model_kwargs={"init_position": 3},
            model_gen_kwargs={"some_gen_kwarg": "some_value"},
        )
        self.branin_experiment = get_branin_experiment()

    def test_minimum_preference_criterion(self) -> None:
        """Tests the minimum preference criterion subclass of TransitionCriterion."""
        criterion = MinimumPreferenceOccurances(metric_name="m1", threshold=3)
        experiment = get_experiment()
        generation_strategy = GenerationStrategy(
            name="SOBOL::default",
            steps=[
                GenerationStep(
                    model=Models.SOBOL,
                    num_trials=-1,
                    completion_criteria=[criterion],
                ),
                GenerationStep(
                    model=Models.GPEI,
                    num_trials=-1,
                    max_parallelism=1,
                ),
            ],
        )
        generation_strategy.experiment = experiment

        # Has not seen enough of each preference
        self.assertFalse(
            generation_strategy._maybe_transition_to_next_node(
                raise_data_required_error=False
            )
        )

        data = Data(
            df=pd.DataFrame(
                {
                    "trial_index": range(6),
                    "arm_name": [f"{i}_0" for i in range(6)],
                    "metric_name": ["m1" for _ in range(6)],
                    "mean": [0, 0, 0, 1, 1, 1],
                    "sem": [0 for _ in range(6)],
                }
            )
        )
        with patch.object(experiment, "fetch_data", return_value=data):
            # We have seen three "yes" and three "no"
            self.assertTrue(
                generation_strategy._maybe_transition_to_next_node(
                    raise_data_required_error=False
                )
            )
            self.assertEqual(
                generation_strategy._curr.model_spec_to_gen_from.model_enum, Models.GPEI
            )

    def test_default_step_criterion_setup(self) -> None:
        """This test ensures that the default completion criterion for GenerationSteps
        is set as expected.

        The default completion criterion is to create two TransitionCriterion, one
        of type `MaximumTrialsInStatus` and one of type `MinTrials`.
        These are constructed via the inputs of `num_trials`, `enforce_num_trials`,
        and `minimum_trials_observed` on the GenerationStep.
        """
        experiment = get_experiment()
        gs = GenerationStrategy(
            name="SOBOL+GPEI::default",
            steps=[
                GenerationStep(
                    model=Models.SOBOL,
                    num_trials=3,
                ),
                GenerationStep(
                    model=Models.GPEI,
                    num_trials=4,
                    max_parallelism=1,
                    min_trials_observed=2,
                    enforce_num_trials=False,
                ),
                GenerationStep(
                    model=Models.GPEI,
                    num_trials=-1,
                ),
            ],
        )
        gs.experiment = experiment

        step_0_expected_transition_criteria = [
            MaxTrials(
                threshold=3,
                block_gen_if_met=True,
                transition_to="GenerationStep_1",
                only_in_statuses=None,
                not_in_statuses=[TrialStatus.FAILED, TrialStatus.ABANDONED],
            ),
        ]
        step_1_expected_transition_criteria = [
            MaxTrials(
                threshold=4,
                block_gen_if_met=False,
                transition_to="GenerationStep_2",
                only_in_statuses=None,
                not_in_statuses=[TrialStatus.FAILED, TrialStatus.ABANDONED],
            ),
            MinTrials(
                only_in_statuses=[TrialStatus.COMPLETED, TrialStatus.EARLY_STOPPED],
                threshold=2,
                transition_to="GenerationStep_2",
            ),
            MaxGenerationParallelism(
                threshold=1,
                only_in_statuses=[TrialStatus.RUNNING],
                block_gen_if_met=True,
                block_transition_if_unmet=False,
            ),
        ]
        step_2_expected_transition_criteria = []
        self.assertEqual(
            gs._steps[0].transition_criteria, step_0_expected_transition_criteria
        )
        self.assertEqual(
            gs._steps[1].transition_criteria, step_1_expected_transition_criteria
        )
        self.assertEqual(
            gs._steps[2].transition_criteria, step_2_expected_transition_criteria
        )

    def test_min_trials_is_met(self) -> None:
        """Test that the is_met method in  MinTrials works"""
        experiment = self.branin_experiment
        gs = GenerationStrategy(
            name="SOBOL::default",
            steps=[
                GenerationStep(
                    model=Models.SOBOL,
                    num_trials=4,
                    min_trials_observed=2,
                    enforce_num_trials=True,
                ),
                GenerationStep(
                    Models.SOBOL,
                    num_trials=-1,
                    max_parallelism=1,
                ),
            ],
        )
        gs.experiment = experiment

        # Need to add trials to test the transition criteria `is_met` method
        for _i in range(4):
            experiment.new_trial(gs.gen(experiment=experiment))

        # TODO: @mgarrard More comprehensive test of trials_from_node
        node_0_trials = gs._steps[0].trials_from_node
        node_1_trials = gs._steps[1].trials_from_node

        self.assertEqual(len(node_0_trials), 4)
        self.assertEqual(len(node_1_trials), 0)

        # MinTrials is met should not pass yet, because no trials
        # are marked completed
        self.assertFalse(
            gs._steps[0]
            .transition_criteria[1]
            .is_met(experiment, gs._steps[0].trials_from_node)
        )

        # Should pass after two trials are marked completed
        for idx, trial in experiment.trials.items():
            trial.mark_running(no_runner_required=True).mark_completed()
            if idx == 1:
                break
        self.assertTrue(
            gs._steps[0]
            .transition_criteria[1]
            .is_met(experiment, gs._steps[0].trials_from_node)
        )

        # Check mixed status MinTrials
        min_criterion = MinTrials(
            threshold=3,
            only_in_statuses=[TrialStatus.COMPLETED, TrialStatus.EARLY_STOPPED],
        )
        self.assertFalse(
            min_criterion.is_met(experiment, gs._steps[0].trials_from_node)
        )
        for idx, trial in experiment.trials.items():
            if idx == 2:
                trial._status = TrialStatus.EARLY_STOPPED
        self.assertTrue(min_criterion.is_met(experiment, gs._steps[0].trials_from_node))

    def test_auto_transition(self) -> None:
        """Very simple test to validate AutoTransitionAfterGenCriterion"""
        experiment = self.branin_experiment
        gs = GenerationStrategy(
            name="test",
            nodes=[
                GenerationNode(
                    node_name="sobol_1",
                    model_specs=[self.sobol_model_spec],
                    transition_criteria=[
                        AutoTransitionAfterGenCriterion(transition_to="sobol_2")
                    ],
                ),
                GenerationNode(
                    node_name="sobol_2", model_specs=[self.sobol_model_spec]
                ),
            ],
        )
        gs.experiment = experiment
        self.assertEqual(gs.current_node_name, "sobol_1")
        gs.gen(experiment=experiment)
        gs.gen(experiment=experiment)
        self.assertEqual(gs.current_node_name, "sobol_2")

    def test_max_trials_is_met(self) -> None:
        """Test that the is_met method in MaxTrials works"""
        experiment = self.branin_experiment
        gs = GenerationStrategy(
            name="SOBOL::default",
            steps=[
                GenerationStep(
                    model=Models.SOBOL,
                    num_trials=4,
                    min_trials_observed=0,
                    enforce_num_trials=True,
                ),
                GenerationStep(
                    model=Models.SOBOL,
                    num_trials=4,
                    min_trials_observed=0,
                    enforce_num_trials=False,
                ),
                GenerationStep(
                    model=Models.SOBOL,
                    num_trials=-1,
                    max_parallelism=1,
                ),
            ],
        )
        gs.experiment = experiment

        # No trials yet, first step should fail
        self.assertFalse(
            gs._steps[0]
            .transition_criteria[0]
            .is_met(
                experiment=experiment,
                trials_from_node=gs._steps[0].trials_from_node,
            )
        )

        # After adding trials, should pass
        for _i in range(4):
            experiment.new_trial(gs.gen(experiment=experiment))
        self.assertTrue(
            gs._steps[0]
            .transition_criteria[0]
            .is_met(
                experiment=experiment,
                trials_from_node=gs._steps[0].trials_from_node,
            )
        )

        # Check not in statuses and only in statuses
        max_criterion_not_in_statuses = MaxTrials(
            threshold=2,
            block_gen_if_met=True,
            not_in_statuses=[TrialStatus.COMPLETED],
        )
        max_criterion_only_statuses = MaxTrials(
            threshold=2,
            block_gen_if_met=True,
            only_in_statuses=[TrialStatus.COMPLETED, TrialStatus.EARLY_STOPPED],
        )
        # experiment currently has 4 trials, but none of them are completed
        self.assertTrue(
            max_criterion_not_in_statuses.is_met(
                experiment, trials_from_node=gs._steps[0].trials_from_node
            )
        )
        self.assertFalse(
            max_criterion_only_statuses.is_met(
                experiment, trials_from_node=gs._steps[0].trials_from_node
            )
        )

        # set 3 of the 4 trials to status == completed
        for _idx, trial in experiment.trials.items():
            trial._status = TrialStatus.COMPLETED
            if _idx == 2:
                break
        self.assertTrue(
            max_criterion_only_statuses.is_met(
                experiment, trials_from_node=gs._steps[0].trials_from_node
            )
        )
        self.assertFalse(
            max_criterion_not_in_statuses.is_met(
                experiment, trials_from_node=gs._steps[0].trials_from_node
            )
        )

    def test_trials_from_node_none(self) -> None:
        """Tests MinTrials and MaxTrials default to experiment
        level trials when trials_from_node is None.
        """
        experiment = get_experiment()
        gs = GenerationStrategy(
            name="SOBOL::default",
            steps=[
                GenerationStep(
                    model=Models.SOBOL,
                    num_trials=4,
                    min_trials_observed=2,
                    enforce_num_trials=True,
                ),
            ],
        )
        max_criterion_with_status = MaxTrials(
            threshold=2,
            block_gen_if_met=True,
            only_in_statuses=[TrialStatus.COMPLETED],
        )
        max_criterion = MaxTrials(threshold=2, block_gen_if_met=True)
        warning_msg = (
            "`trials_from_node` is None, will check threshold on" + " experiment level."
        )

        # no trials so criterion should be false, then add trials to pass criterion
        with self.assertLogs(TransitionCriterion.__module__, logging.WARNING) as logger:
            self.assertFalse(max_criterion.is_met(experiment, trials_from_node=None))
            self.assertTrue(
                any(warning_msg in output for output in logger.output),
                logger.output,
            )
        for _i in range(3):
            experiment.new_trial(gs.gen(experiment=experiment))
        self.assertTrue(max_criterion.is_met(experiment, trials_from_node=None))

        # Before marking trial status it should be false, until trials are completed
        self.assertFalse(
            max_criterion_with_status.is_met(experiment, trials_from_node=None)
        )
        for idx, trial in experiment.trials.items():
            trial._status = TrialStatus.COMPLETED
            if idx == 1:
                break
        self.assertTrue(
            max_criterion_with_status.is_met(experiment, trials_from_node=None)
        )

        # Check MinTrials
        min_criterion = MinTrials(
            threshold=3,
            only_in_statuses=[TrialStatus.COMPLETED, TrialStatus.EARLY_STOPPED],
        )
        with self.assertLogs(TransitionCriterion.__module__, logging.WARNING) as logger:
            self.assertFalse(min_criterion.is_met(experiment, trials_from_node=None))
            self.assertTrue(
                any(warning_msg in output for output in logger.output),
                logger.output,
            )
        for _idx, trial in experiment.trials.items():
            trial._status = TrialStatus.COMPLETED
        self.assertTrue(min_criterion.is_met(experiment, trials_from_node=None))

    def test_repr(self) -> None:
        """Tests that the repr string is correctly formatted for all
        TransitionCriterion child classes.
        """
        self.maxDiff = None
        max_trials_criterion = MaxTrials(
            threshold=5,
            block_gen_if_met=True,
            block_transition_if_unmet=False,
            transition_to="GenerationStep_1",
            only_in_statuses=[TrialStatus.COMPLETED],
            not_in_statuses=[TrialStatus.FAILED],
        )
        self.assertEqual(
            str(max_trials_criterion),
            "MaxTrials({'threshold': 5, "
            + "'only_in_statuses': [<enum 'TrialStatus'>.COMPLETED], "
            + "'not_in_statuses': [<enum 'TrialStatus'>.FAILED], "
            + "'transition_to': 'GenerationStep_1', "
            + "'block_transition_if_unmet': False, "
            + "'block_gen_if_met': True, "
            + "'use_all_trials_in_exp': False, "
            + "'continue_trial_generation': False})",
        )
        minimum_trials_in_status_criterion = MinTrials(
            only_in_statuses=[TrialStatus.COMPLETED, TrialStatus.EARLY_STOPPED],
            threshold=0,
            transition_to="GenerationStep_2",
            block_gen_if_met=True,
            block_transition_if_unmet=False,
            not_in_statuses=[TrialStatus.FAILED],
        )
        self.assertEqual(
            str(minimum_trials_in_status_criterion),
            "MinTrials({'threshold': 0, 'only_in_statuses': "
            + "[<enum 'TrialStatus'>.COMPLETED, <enum 'TrialStatus'>.EARLY_STOPPED], "
            + "'not_in_statuses': [<enum 'TrialStatus'>.FAILED], "
            + "'transition_to': 'GenerationStep_2', "
            + "'block_transition_if_unmet': False, "
            + "'block_gen_if_met': True, "
            + "'use_all_trials_in_exp': False, "
            + "'continue_trial_generation': False})",
        )
        minimum_preference_occurrences_criterion = MinimumPreferenceOccurances(
            metric_name="m1", threshold=3
        )
        self.assertEqual(
            str(minimum_preference_occurrences_criterion),
            "MinimumPreferenceOccurances({'metric_name': 'm1', 'threshold': 3, "
            + "'transition_to': None, 'block_gen_if_met': False, "
            "'block_transition_if_unmet': True})",
        )
        deprecated_min_trials_criterion = MinimumTrialsInStatus(
            status=TrialStatus.COMPLETED, threshold=3
        )
        self.assertEqual(
            str(deprecated_min_trials_criterion),
            "MinimumTrialsInStatus({"
            + "'status': <enum 'TrialStatus'>.COMPLETED, "
            + "'threshold': 3, "
            + "'transition_to': None})",
        )
        max_parallelism = MaxGenerationParallelism(
            only_in_statuses=[TrialStatus.EARLY_STOPPED],
            threshold=3,
            transition_to="GenerationStep_2",
            block_gen_if_met=True,
            block_transition_if_unmet=False,
            not_in_statuses=[TrialStatus.FAILED],
        )
        self.assertEqual(
            str(max_parallelism),
            "MaxGenerationParallelism({'threshold': 3, 'only_in_statuses': "
            + "[<enum 'TrialStatus'>.EARLY_STOPPED], "
            + "'not_in_statuses': [<enum 'TrialStatus'>.FAILED], "
            + "'transition_to': 'GenerationStep_2', "
            + "'block_transition_if_unmet': False, "
            + "'block_gen_if_met': True, "
            + "'use_all_trials_in_exp': False, "
            + "'continue_trial_generation': True})",
        )
        auto_transition = AutoTransitionAfterGenCriterion(
            transition_to="GenerationStep_2"
        )
        self.assertEqual(
            str(auto_transition),
            "AutoTransitionAfterGenCriterion({'transition_to': 'GenerationStep_2', "
            + "'block_transition_if_unmet': True, "
            + "'continue_trial_generation': True})",
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import copy

import plotly.graph_objects as go
from ax.analysis.cross_validation_plot import CrossValidationPlot
from ax.analysis.helpers.cross_validation_helpers import (
    error_scatter_data_from_cv_results,
)
from ax.analysis.helpers.scatter_helpers import error_scatter_trace_from_df
from ax.modelbridge.cross_validation import cross_validate
from ax.modelbridge.registry import Models
from ax.plot.base import PlotMetric
from ax.plot.diagnostic import (
    _get_cv_plot_data as PLOT_get_cv_plot_data,
    interact_cross_validation_plotly as PLOT_interact_cross_validation_plotly,
)
from ax.plot.scatter import (
    _error_scatter_data as PLOT_error_scatter_data,
    _error_scatter_trace as PLOT_error_scatter_trace,
)
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment
from ax.utils.testing.mock import fast_botorch_optimize


class TestCVConsistencyCheck(TestCase):
    @fast_botorch_optimize
    def setUp(self) -> None:
        super().setUp()
        self.exp = get_branin_experiment(with_batch=True)
        self.exp.trials[0].run()
        self.model = Models.BOTORCH_MODULAR(
            # Model bridge kwargs
            experiment=self.exp,
            data=self.exp.fetch_data(),
        )

        self.exp_status_quo = get_branin_experiment(
            with_batch=True, with_status_quo=True
        )
        self.exp_status_quo.trials[0].run()
        self.model_status_quo = Models.BOTORCH_MODULAR(
            # Model bridge kwargs
            experiment=self.exp,
            data=self.exp.fetch_data(),
        )

    def test_error_scatter_data_branin(self) -> None:
        cv_results = cross_validate(self.model)
        cv_results_plot = copy.deepcopy(cv_results)

        result_analysis = error_scatter_data_from_cv_results(
            cv_results=cv_results,
            metric_name="branin",
        )

        data = PLOT_get_cv_plot_data(cv_results_plot, label_dict={})
        result_plot = PLOT_error_scatter_data(
            list(data.in_sample.values()),
            y_axis_var=PlotMetric("branin", pred=True, rel=False),
            x_axis_var=PlotMetric("branin", pred=False, rel=False),
        )
        self.assertEqual(result_analysis, result_plot)

    def test_error_scatter_trace_branin(self) -> None:
        cv_results = cross_validate(self.model)
        cv_results_plot = copy.deepcopy(cv_results)

        cross_validation_plot = CrossValidationPlot(
            experiment=self.exp, model=self.model
        )
        df = cross_validation_plot.get_df()

        metric_filtered_df = df.loc[df["metric_name"] == "branin"]
        result_analysis = error_scatter_trace_from_df(
            df=metric_filtered_df,
            show_CI=True,
            visible=True,
            x_axis_label="Actual Outcome",
            y_axis_label="Predicted Outcome",
        )

        data = data = PLOT_get_cv_plot_data(cv_results_plot, label_dict={})

        result_plot = PLOT_error_scatter_trace(
            arms=list(data.in_sample.values()),
            hoverinfo="text",
            show_arm_details_on_hover=True,
            show_CI=True,
            show_context=False,
            status_quo_arm=None,
            visible=True,
            y_axis_var=PlotMetric("branin", pred=True, rel=False),
            x_axis_var=PlotMetric("branin", pred=False, rel=False),
            x_axis_label="Actual Outcome",
            y_axis_label="Predicted Outcome",
        )

        print(str(result_analysis))
        print(str(result_plot))

        self.assertEqual(result_analysis, result_plot)

    def test_obs_vs_pred_dropdown_plot_branin(self) -> None:

        label_dict = {"branin": "BrAnIn"}

        cross_validation_plot = CrossValidationPlot(
            experiment=self.exp, model=self.model, label_dict=label_dict
        )
        fig = cross_validation_plot.get_fig()

        self.assertIsInstance(fig, go.Figure)

        cv_results_plot = cross_validate(self.model)

        fig_PLOT = PLOT_interact_cross_validation_plotly(
            cv_results_plot,
            show_context=False,
            label_dict=label_dict,
            caption=CrossValidationPlot.CROSS_VALIDATION_CAPTION,
        )
        self.assertEqual(fig, fig_PLOT)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import random
from typing import List
from unittest.mock import MagicMock, patch

import pandas as pd
import torch
from ax.core.arm import Arm
from ax.core.batch_trial import BatchTrial
from ax.core.data import Data
from ax.core.generator_run import GeneratorRun
from ax.core.objective import ScalarizedObjective
from ax.core.optimization_config import OptimizationConfig
from ax.core.outcome_constraint import OutcomeConstraint
from ax.core.types import ComparisonOp
from ax.exceptions.core import UserInputError
from ax.modelbridge.cross_validation import AssessModelFitResult
from ax.modelbridge.registry import Models
from ax.modelbridge.torch import TorchModelBridge
from ax.plot.pareto_utils import get_tensor_converter_model
from ax.service.utils.best_point import (
    _derel_opt_config_wrapper,
    _is_row_feasible,
    extract_Y_from_data,
    get_best_parameters,
    get_best_raw_objective_point,
    logger as best_point_logger,
)
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import not_none
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_branin_metric,
    get_experiment_with_observations,
    get_sobol,
)
from ax.utils.testing.mock import fast_botorch_optimize

best_point_module: str = _derel_opt_config_wrapper.__module__
DUMMY_OPTIMIZATION_CONFIG = "test_optimization_config"


class TestBestPointUtils(TestCase):
    """Testing the best point utilities functionality that is not tested in
    main `AxClient` testing suite (`TestServiceAPI`)."""

    @fast_botorch_optimize
    def test_best_from_model_prediction(self) -> None:
        exp = get_branin_experiment()

        for _ in range(3):
            sobol = Models.SOBOL(search_space=exp.search_space)
            generator_run = sobol.gen(n=1)
            trial = exp.new_trial(generator_run=generator_run)
            trial.run()
            trial.mark_completed()
            exp.attach_data(exp.fetch_data())

        model = Models.BOTORCH_MODULAR(experiment=exp, data=exp.lookup_data())
        generator_run = model.gen(n=1)
        trial = exp.new_trial(generator_run=generator_run)
        trial.run()
        trial.mark_completed()

        with patch.object(
            TorchModelBridge,
            "model_best_point",
            return_value=(
                (
                    Arm(
                        name="0_0",
                        parameters={"x1": -4.842811906710267, "x2": 11.887089014053345},
                    ),
                    (
                        {"branin": 34.76260622783635},
                        {"branin": {"branin": 0.00028306433439807734}},
                    ),
                )
            ),
        ) as mock_model_best_point, self.assertLogs(
            logger=best_point_logger, level="WARN"
        ) as lg:
            # Test bad model fit causes function to resort back to raw data
            with patch(
                "ax.service.utils.best_point.assess_model_fit",
                return_value=AssessModelFitResult(
                    good_fit_metrics_to_fisher_score={},
                    bad_fit_metrics_to_fisher_score={
                        "branin": 0,
                    },
                ),
            ):
                self.assertIsNotNone(get_best_parameters(exp, Models))
                self.assertTrue(
                    any("Model fit is poor" in warning for warning in lg.output),
                    msg=lg.output,
                )
                mock_model_best_point.assert_not_called()

            # Test model best point is used when fit is good
            with patch(
                "ax.service.utils.best_point.assess_model_fit",
                return_value=AssessModelFitResult(
                    good_fit_metrics_to_fisher_score={
                        "branin": 0,
                    },
                    bad_fit_metrics_to_fisher_score={},
                ),
            ):
                self.assertIsNotNone(get_best_parameters(exp, Models))
                mock_model_best_point.assert_called()

        # Assert the non-mocked method works correctly as well
        self.assertIsNotNone(get_best_parameters(exp, Models))

    def test_best_raw_objective_point(self) -> None:
        exp = get_branin_experiment()
        with self.assertRaisesRegex(ValueError, "Cannot identify best "):
            get_best_raw_objective_point(exp)
        self.assertEqual(get_best_parameters(exp, Models), None)
        exp.new_trial(
            generator_run=GeneratorRun(arms=[Arm(parameters={"x1": 5.0, "x2": 5.0})])
        ).run().complete()
        exp.fetch_data()
        # pyre-fixme[16]: Optional type has no attribute `clone`.
        opt_conf = exp.optimization_config.clone()
        opt_conf.objective.metric._name = "not_branin"
        with self.assertRaisesRegex(ValueError, "No data has been logged"):
            get_best_raw_objective_point(exp, opt_conf)

        # Test constraints work as expected.
        observations = [[1.0, 2.0], [3.0, 4.0], [-5.0, -6.0]]
        exp = get_experiment_with_observations(
            observations=observations,
            constrained=True,
            minimize=False,
        )
        _, best_prediction = not_none(get_best_parameters(exp, Models))
        best_metrics = not_none(best_prediction)[0]
        self.assertDictEqual(best_metrics, {"m1": 3.0, "m2": 4.0})

        # Tensor bounds are accepted.
        constraint = not_none(exp.optimization_config).all_constraints[0]
        # pyre-fixme[8]: Attribute `bound` declared in class `OutcomeConstraint`
        # has type `float` but is used as type `Tensor`.
        constraint.bound = torch.tensor(constraint.bound)
        _, best_prediction = not_none(get_best_parameters(exp, Models))
        best_metrics = not_none(best_prediction)[0]
        self.assertDictEqual(best_metrics, {"m1": 3.0, "m2": 4.0})

    def test_best_raw_objective_point_unsatisfiable(self) -> None:
        exp = get_branin_experiment()
        trial = exp.new_trial(
            generator_run=GeneratorRun(arms=[Arm(parameters={"x1": 5.0, "x2": 5.0})])
        ).run()
        trial.mark_completed()
        exp.fetch_data()

        # pyre-fixme[16]: Optional type has no attribute `clone`.
        opt_conf = exp.optimization_config.clone()
        opt_conf.outcome_constraints.append(
            OutcomeConstraint(
                metric=get_branin_metric(), op=ComparisonOp.LEQ, bound=0, relative=False
            )
        )

        with self.assertRaisesRegex(ValueError, "No points satisfied"):
            get_best_raw_objective_point(exp, opt_conf)

    def test_best_raw_objective_point_unsatisfiable_relative(self) -> None:
        exp = get_experiment_with_observations(
            observations=[[-1, 1]],
            constrained=True,
        )

        # Create altered optimization config with unsatisfiable relative constraint.
        opt_conf = not_none(exp.optimization_config).clone()
        opt_conf.outcome_constraints[0].relative = True
        opt_conf.outcome_constraints[0].bound = 9999

        with self.assertLogs(logger=best_point_logger, level="WARN") as lg:
            get_best_raw_objective_point(exp, opt_conf)
            self.assertTrue(
                any("No status quo provided" in warning for warning in lg.output),
                msg=lg.output,
            )

        exp.status_quo = exp.trials[0].arms[0]

        with self.assertRaisesRegex(ValueError, "No points satisfied"):
            get_best_raw_objective_point(exp, opt_conf)

    def test_best_raw_objective_point_scalarized(self) -> None:
        exp = get_branin_experiment()
        exp.optimization_config = OptimizationConfig(
            ScalarizedObjective(metrics=[get_branin_metric()], minimize=True)
        )
        with self.assertRaisesRegex(ValueError, "Cannot identify best "):
            get_best_raw_objective_point(exp)
        self.assertEqual(get_best_parameters(exp, Models), None)
        exp.new_trial(
            generator_run=GeneratorRun(arms=[Arm(parameters={"x1": 5.0, "x2": 5.0})])
        ).run().complete()
        exp.fetch_data()
        self.assertEqual(get_best_raw_objective_point(exp)[0], {"x1": 5.0, "x2": 5.0})

    def test_best_raw_objective_point_scalarized_multi(self) -> None:
        exp = get_branin_experiment()
        exp.optimization_config = OptimizationConfig(
            ScalarizedObjective(
                metrics=[get_branin_metric(), get_branin_metric(lower_is_better=False)],
                weights=[0.1, -0.9],
                minimize=True,
            )
        )
        with self.assertRaisesRegex(ValueError, "Cannot identify best "):
            get_best_raw_objective_point(exp)
        self.assertEqual(get_best_parameters(exp, Models), None)
        exp.new_trial(
            generator_run=GeneratorRun(arms=[Arm(parameters={"x1": 5.0, "x2": 5.0})])
        ).run().complete()
        exp.fetch_data()
        self.assertEqual(get_best_raw_objective_point(exp)[0], {"x1": 5.0, "x2": 5.0})

    @patch(
        f"{best_point_module}.derelativize_optimization_config_with_raw_status_quo",
        return_value=DUMMY_OPTIMIZATION_CONFIG,
    )
    def test_derel_opt_config_wrapper(self, mock_derelativize: MagicMock) -> None:
        # No change to optimization config without relative constraints/thresholds.
        exp = get_experiment_with_observations(
            observations=[[-1, 1, 1], [1, 2, 1], [3, 3, -1], [2, 4, 1], [2, 0, 1]],
            constrained=True,
        )
        input_optimization_config = not_none(exp.optimization_config)
        optimization_config = _derel_opt_config_wrapper(
            optimization_config=input_optimization_config
        )
        self.assertEqual(input_optimization_config, optimization_config)

        # Add relative constraints.
        for constraint in input_optimization_config.all_constraints:
            constraint.relative = True

        # Check errors.
        with self.assertRaisesRegex(
            ValueError,
            "Must specify ModelBridge or Experiment when calling "
            "`_derel_opt_config_wrapper`.",
        ):
            _derel_opt_config_wrapper(optimization_config=input_optimization_config)
        with self.assertRaisesRegex(
            ValueError,
            "`modelbridge` must have status quo if specified. If `modelbridge` is "
            "unspecified, `experiment` must have a status quo.",
        ):
            _derel_opt_config_wrapper(
                optimization_config=input_optimization_config, experiment=exp
            )

        # Set status quo.
        exp.status_quo = exp.trials[0].arms[0]

        # ModelBridges will have specific addresses and so must be self-same to
        # pass equality checks.
        test_modelbridge_1 = get_tensor_converter_model(
            experiment=not_none(exp),
            data=not_none(exp).lookup_data(),
        )
        test_observations_1 = test_modelbridge_1.get_training_data()
        returned_value = _derel_opt_config_wrapper(
            optimization_config=input_optimization_config,
            modelbridge=test_modelbridge_1,
            observations=test_observations_1,
        )
        mock_derelativize.assert_called_with(
            optimization_config=input_optimization_config,
            modelbridge=test_modelbridge_1,
            observations=test_observations_1,
        )
        with patch(
            f"{best_point_module}.get_tensor_converter_model",
            return_value=test_modelbridge_1,
        ), patch(
            f"{best_point_module}.ModelBridge.get_training_data",
            return_value=test_observations_1,
        ):
            returned_value = _derel_opt_config_wrapper(
                optimization_config=input_optimization_config, experiment=exp
            )
        self.assertEqual(returned_value, DUMMY_OPTIMIZATION_CONFIG)
        mock_derelativize.assert_called_with(
            optimization_config=input_optimization_config,
            modelbridge=test_modelbridge_1,
            observations=test_observations_1,
        )

        # Observations and ModelBridge are not constructed from other inputs when
        # provided.
        test_modelbridge_2 = get_tensor_converter_model(
            experiment=not_none(exp),
            data=not_none(exp).lookup_data(),
        )
        test_observations_2 = test_modelbridge_2.get_training_data()
        with self.assertLogs(logger=best_point_logger, level="WARN") as lg, patch(
            f"{best_point_module}.get_tensor_converter_model",
            return_value=test_modelbridge_2,
        ), patch(
            f"{best_point_module}.ModelBridge.get_training_data",
            return_value=test_observations_2,
        ):
            returned_value = _derel_opt_config_wrapper(
                optimization_config=input_optimization_config,
                experiment=exp,
                modelbridge=test_modelbridge_1,
                observations=test_observations_1,
            )
            self.assertTrue(
                any(
                    "ModelBridge and Experiment provided to "
                    "`_derel_opt_config_wrapper`. Ignoring the latter." in warning
                    for warning in lg.output
                ),
                msg=lg.output,
            )
        self.assertEqual(returned_value, DUMMY_OPTIMIZATION_CONFIG)
        mock_derelativize.assert_called_with(
            optimization_config=input_optimization_config,
            modelbridge=test_modelbridge_1,
            observations=test_observations_1,
        )

    def test_extract_Y_from_data(self) -> None:
        experiment = get_branin_experiment()
        sobol_generator = get_sobol(search_space=experiment.search_space)
        for i in range(20):
            sobol_run = sobol_generator.gen(n=1)
            trial = experiment.new_trial(generator_run=sobol_run).mark_running(
                no_runner_required=True
            )
            if i in [3, 8, 10]:
                trial.mark_early_stopped()
            else:
                trial.mark_completed()

        df_dicts = []
        for trial_idx in range(20):
            for metric_name in ["foo", "bar"]:
                df_dicts.append(
                    {
                        "trial_index": trial_idx,
                        "metric_name": metric_name,
                        "arm_name": f"{trial_idx}_0",
                        "mean": (
                            float(trial_idx)
                            if metric_name == "foo"
                            else trial_idx + 5.0
                        ),
                        "sem": 0.0,
                    }
                )
        experiment.attach_data(Data(df=pd.DataFrame.from_records(df_dicts)))

        expected_Y = torch.stack(
            [
                torch.arange(20, dtype=torch.double),
                torch.arange(5, 25, dtype=torch.double),
            ],
            dim=-1,
        )
        Y, trial_indices = extract_Y_from_data(
            experiment=experiment,
            metric_names=["foo", "bar"],
        )
        expected_trial_indices = torch.arange(20)
        self.assertTrue(torch.allclose(Y, expected_Y))
        self.assertTrue(torch.equal(trial_indices, expected_trial_indices))
        # Check that it respects ordering of metric names.
        Y, trial_indices = extract_Y_from_data(
            experiment=experiment,
            metric_names=["bar", "foo"],
        )
        self.assertTrue(torch.allclose(Y, expected_Y[:, [1, 0]]))
        self.assertTrue(torch.equal(trial_indices, expected_trial_indices))
        # Extract partial metrics.
        Y, trial_indices = extract_Y_from_data(
            experiment=experiment, metric_names=["bar"]
        )
        self.assertTrue(torch.allclose(Y, expected_Y[:, [1]]))
        self.assertTrue(torch.equal(trial_indices, expected_trial_indices))
        # Works with messed up ordering of data.
        clone_dicts = df_dicts.copy()
        random.shuffle(clone_dicts)
        experiment._data_by_trial = {}
        experiment.attach_data(Data(df=pd.DataFrame.from_records(clone_dicts)))
        Y, trial_indices = extract_Y_from_data(
            experiment=experiment,
            metric_names=["foo", "bar"],
        )
        self.assertTrue(torch.allclose(Y, expected_Y))
        self.assertTrue(torch.equal(trial_indices, expected_trial_indices))

        # Check that it skips trials that are not completed.
        experiment.trials[0].mark_running(no_runner_required=True, unsafe=True)
        experiment.trials[1].mark_abandoned(unsafe=True)
        Y, trial_indices = extract_Y_from_data(
            experiment=experiment,
            metric_names=["foo", "bar"],
        )
        self.assertTrue(torch.allclose(Y, expected_Y[2:]))
        self.assertTrue(torch.equal(trial_indices, expected_trial_indices[2:]))

        # Error with missing data.
        with self.assertRaisesRegex(
            UserInputError, "Trial 2 is missing data on metrics {'foo'}."
        ):
            # Skipping first 5 data points since first two trials are not completed.

            extract_Y_from_data(
                experiment=experiment,
                metric_names=["foo", "bar"],
                data=Data(df=pd.DataFrame.from_records(df_dicts[5:])),
            )

        # Error with extra data.
        with self.assertRaisesRegex(
            UserInputError, "Trial data has more than one row per arm, metric pair. "
        ):
            # Skipping first 5 data points since first two trials are not completed.
            base_df = pd.DataFrame.from_records(df_dicts[5:])
            extract_Y_from_data(
                experiment=experiment,
                metric_names=["foo", "bar"],
                data=Data(df=pd.concat((base_df, base_df))),
            )

        # Check that it works with BatchTrial.
        experiment = get_branin_experiment()
        batch_trial = BatchTrial(experiment=experiment, index=0)
        batch_trial.add_arm(Arm(name="0_0", parameters={"x1": 0.0, "x2": 0.0}))
        batch_trial.add_arm(Arm(name="0_1", parameters={"x1": 1.0, "x2": 0.0}))
        batch_trial.mark_running(no_runner_required=True).mark_completed()
        df_dicts_batch = []
        for i in (0, 1):
            for metric_name in ["foo", "bar"]:
                df_dicts_batch.append(
                    {
                        "trial_index": 0,
                        "metric_name": metric_name,
                        "arm_name": f"0_{i}",
                        "mean": float(i) if metric_name == "foo" else i + 5.0,
                        "sem": 0.0,
                    }
                )
        batch_df = pd.DataFrame.from_records(df_dicts_batch)
        Y, trial_indices = extract_Y_from_data(
            experiment=experiment,
            metric_names=["foo", "bar"],
            data=Data(df=batch_df),
        )
        self.assertTrue(torch.allclose(Y, expected_Y[:2]))
        self.assertTrue(torch.equal(trial_indices, torch.zeros(2, dtype=torch.long)))

    def test_is_row_feasible(self) -> None:
        exp = get_experiment_with_observations(
            observations=[[-1, 1, 1], [1, 2, 1], [3, 3, -1], [2, 4, 1], [2, 0, 1]],
            constrained=True,
        )
        feasible_series = _is_row_feasible(
            df=exp.lookup_data().df,
            optimization_config=not_none(exp.optimization_config),
        )
        expected_per_arm = [False, True, False, True, True]
        expected_series = _repeat_elements(
            list_to_replicate=expected_per_arm, n_repeats=3
        )
        pd.testing.assert_series_equal(
            feasible_series, expected_series, check_names=False
        )

        exp.optimization_config.outcome_constraints[0].relative = True
        relative_constraint_warning = (
            "WARNING:ax.service.utils.best_point:Ignoring relative constraint "
            "OutcomeConstraint(m3 >= 0.0%). Derelativize OptimizationConfig "
            "before passing to `_is_row_feasible`."
        )
        with self.assertLogs(logger=best_point_logger, level="WARN") as lg:
            # with lookout for warnings(" OutcomeConstraint(m3 >= 0.0%) ignored."):
            feasible_series = _is_row_feasible(
                df=exp.lookup_data().df,
                optimization_config=not_none(exp.optimization_config),
            )
            self.assertTrue(
                any(relative_constraint_warning in warning for warning in lg.output),
                msg=lg.output,
            )
        expected_per_arm = [False, True, True, True, True]
        expected_series = _repeat_elements(
            list_to_replicate=expected_per_arm, n_repeats=3
        )
        pd.testing.assert_series_equal(
            feasible_series, expected_series, check_names=False
        )
        exp._status_quo = exp.trials[0].arms[0]
        for constraint in not_none(exp.optimization_config).all_constraints:
            constraint.relative = True
        optimization_config = _derel_opt_config_wrapper(
            optimization_config=not_none(exp.optimization_config),
            experiment=exp,
        )
        with self.assertLogs(logger=best_point_logger, level="WARN") as lg:
            # `assertNoLogs` coming in 3.10 - until then we log a dummy warning and
            # continue.
            best_point_logger.warning("Dummy warning")
            feasible_series = _is_row_feasible(
                df=exp.lookup_data().df, optimization_config=optimization_config
            )
            self.assertFalse(
                any(relative_constraint_warning in warning for warning in lg.output),
                msg=lg.output,
            )
        expected_per_arm = [True, True, False, True, False]
        expected_series = _repeat_elements(
            list_to_replicate=expected_per_arm, n_repeats=3
        )
        pd.testing.assert_series_equal(
            feasible_series, expected_series, check_names=False
        )

        # Check that index is carried over for interfacing appropriately
        # with related dataframes.
        exp = get_experiment_with_observations(
            observations=[[-1, 1, 1], [1, 2, 1], [3, 3, -1], [2, 4, 1], [2, 0, 1]],
            constrained=False,
        )
        df = exp.lookup_data().df
        # Artificially redact some data.
        df = df[df["mean"] > 1]
        feasible_series = _is_row_feasible(
            df=df, optimization_config=optimization_config
        )
        pd.testing.assert_index_equal(
            df.index, feasible_series.index, check_names=False
        )


def _repeat_elements(list_to_replicate: List[bool], n_repeats: int) -> pd.Series:
    return pd.Series([item for item in list_to_replicate for _ in range(n_repeats)])



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy

import numpy as np
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.parameter import ChoiceParameter, ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.modelbridge.transforms.map_unit_x import MapUnitX
from ax.utils.common.testutils import TestCase


class MapUnitXTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.target_lb = MapUnitX.target_lb
        self.target_range = MapUnitX.target_range
        self.target_ub = self.target_lb + self.target_range
        self.search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    "x", lower=1, upper=3, parameter_type=ParameterType.FLOAT
                ),
                RangeParameter("a", lower=1, upper=2, parameter_type=ParameterType.INT),
                ChoiceParameter(
                    "b", parameter_type=ParameterType.STRING, values=["a", "b", "c"]
                ),
            ],
        )
        self.observation_features = [
            ObservationFeatures(
                parameters={"x": 2, "a": 2, "b": "b", "step_1": 0.0, "step_2": 0.0}
            ),
            ObservationFeatures(
                parameters={"x": 2, "a": 2, "b": "b", "step_1": 10.0, "step_2": 20.0}
            ),
            ObservationFeatures(parameters={"x": 2, "a": 2, "b": "b", "step_2": 15.0}),
            ObservationFeatures(
                parameters={"x": 2, "a": 2, "b": "b", "step_1": 2.0, "step_2": 12.0}
            ),
            ObservationFeatures(parameters={"x": 2, "a": 2, "b": "b", "step_1": 3.0}),
            ObservationFeatures(
                parameters={"x": 2, "a": 2, "b": "b", "step_1": 7.0, "step_2": 3.0}
            ),
        ]
        self.observations = [
            Observation(
                data=ObservationData([], np.array([]), np.empty((0, 0))), features=obsf
            )
            for obsf in self.observation_features
        ]
        self.t = MapUnitX(
            search_space=self.search_space,
            observations=self.observations,
        )

    def test_Init(self) -> None:
        self.assertEqual(self.t.bounds, {"step_1": (0.0, 10.0), "step_2": (0.0, 20.0)})

    def test_TransformObservationFeatures(self) -> None:
        obs_ft2 = deepcopy(self.observation_features)
        obs_ft2 = self.t.transform_observation_features(obs_ft2)
        expected = [
            ObservationFeatures(
                parameters=self._construct_expected_param_dict(obsf, self.t)
            )
            for obsf in self.observation_features
        ]
        for obsf, expected_obsf in zip(obs_ft2, expected):
            for step in ("step_1", "step_2"):
                if step in expected_obsf.parameters:
                    self.assertAlmostEqual(
                        obsf.parameters[step],
                        # pyre-fixme[6]: For 2nd argument expected
                        #  `SupportsRSub[Variable[_T],
                        #  SupportsAbs[SupportsRound[object]]]` but got `Union[None,
                        #  bool, float, int, str]`.
                        expected_obsf.parameters[step],
                    )

        obs_ft2 = self.t.untransform_observation_features(obs_ft2)
        for obsf, expected_obsf in zip(obs_ft2, self.observation_features):
            for step in ("step_1", "step_2"):
                if step in expected_obsf.parameters:
                    self.assertAlmostEqual(
                        obsf.parameters[step],
                        # pyre-fixme[6]: For 2nd argument expected
                        #  `SupportsRSub[Variable[_T],
                        #  SupportsAbs[SupportsRound[object]]]` but got `Union[None,
                        #  bool, float, int, str]`.
                        expected_obsf.parameters[step],
                    )

    def test_TransformSearchSpace(self) -> None:
        ss2 = deepcopy(self.search_space)
        ss2 = self.t.transform_search_space(ss2)
        self.assertEqual(ss2.parameters, self.search_space.parameters)

    # pyre-fixme[3]: Return type must be annotated.
    def _construct_expected_param_dict(
        self, obsf: ObservationFeatures, transform: MapUnitX
    ):
        result = {"x": 2, "a": 2, "b": "b"}
        for step in ("step_1", "step_2"):
            if step in obsf.parameters:
                scale_fac = self.target_range / (
                    transform.bounds[step][1] - transform.bounds[step][0]
                )
                # pyre-fixme[6]: For 2nd param expected `Union[int, str]` but got
                #  `float`.
                # pyre-fixme[58]: `*` is not supported for operand types
                #  `Union[None, bool, float, int, str]` and `float`.
                result[step] = self.target_lb + obsf.parameters[step] * scale_fac
        return result



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.metrics.branin import BraninMetric
from ax.runners.synthetic import SyntheticRunner
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_arms, get_multi_type_experiment


class MultiTypeExperimentTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.experiment = get_multi_type_experiment()

    def test_MTExperimentFlow(self) -> None:
        self.assertTrue(self.experiment.supports_trial_type("type1"))
        self.assertTrue(self.experiment.supports_trial_type("type2"))
        self.assertFalse(self.experiment.supports_trial_type(None))

        n = 10
        arms = get_branin_arms(n=n, seed=0)

        b1 = self.experiment.new_batch_trial()
        b1.add_arms_and_weights(arms=arms)
        self.assertEqual(b1.trial_type, "type1")
        b1.run()
        self.assertEqual(b1.run_metadata["dummy_metadata"], "dummy1")

        self.experiment.update_runner("type2", SyntheticRunner(dummy_metadata="dummy3"))
        b2 = self.experiment.new_batch_trial()
        b2.trial_type = "type2"
        b2.add_arms_and_weights(arms=arms)
        self.assertEqual(b2.trial_type, "type2")
        b2.run()
        self.assertEqual(b2.run_metadata["dummy_metadata"], "dummy3")

        df = self.experiment.fetch_data().df
        for _, row in df.iterrows():
            # Make sure proper metric present for each batch only
            self.assertEqual(
                row["metric_name"], "m1" if row["trial_index"] == 0 else "m2"
            )

        arm_0_slice = df.loc[df["arm_name"] == "0_0"]
        self.assertNotEqual(
            float(arm_0_slice[df["trial_index"] == 0]["mean"]),
            float(arm_0_slice[df["trial_index"] == 1]["mean"]),
        )
        self.assertEqual(len(df), 2 * n)
        self.assertEqual(self.experiment.default_trials, {0})
        # Set 2 metrics to be equal
        self.experiment.update_tracking_metric(
            BraninMetric("m2", ["x1", "x2"]), trial_type="type2"
        )
        df = self.experiment.fetch_data().df
        arm_0_slice = df.loc[df["arm_name"] == "0_0"]
        self.assertAlmostEqual(
            float(arm_0_slice[df["trial_index"] == 0]["mean"]),
            float(arm_0_slice[df["trial_index"] == 1]["mean"]),
            places=10,
        )

    def test_Repr(self) -> None:
        self.assertEqual(str(self.experiment), "MultiTypeExperiment(test_exp)")

    def test_Eq(self) -> None:
        exp2 = get_multi_type_experiment()

        # Should be equal to start
        self.assertTrue(self.experiment == exp2)

        self.experiment.add_tracking_metric(
            BraninMetric("m3", ["x2", "x1"]), trial_type="type1", canonical_name="m4"
        )

        # Test different set of metrics
        self.assertFalse(self.experiment == exp2)

        exp2.add_tracking_metric(
            BraninMetric("m3", ["x2", "x1"]), trial_type="type1", canonical_name="m5"
        )

        # Test different metric definitions
        self.assertFalse(self.experiment == exp2)

        exp2.update_tracking_metric(
            BraninMetric("m3", ["x2", "x1"]), trial_type="type1", canonical_name="m4"
        )

        # Should be the same
        self.assertTrue(self.experiment == exp2)

        exp2.remove_tracking_metric("m3")
        self.assertFalse(self.experiment == exp2)

    def test_BadBehavior(self) -> None:
        # Add trial type that already exists
        with self.assertRaises(ValueError):
            self.experiment.add_trial_type("type1", SyntheticRunner())

        # Update runner for non-existent trial type
        with self.assertRaises(ValueError):
            self.experiment.update_runner("type3", SyntheticRunner())

        # Add metric for trial_type that doesn't exist
        with self.assertRaises(ValueError):
            self.experiment.add_tracking_metric(
                BraninMetric("m2", ["x1", "x2"]), "type3"
            )

        # Try to remove metric that doesn't exist
        with self.assertRaises(ValueError):
            self.experiment.remove_tracking_metric("m3")

        # Try to change optimization metric to non-primary trial type
        with self.assertRaises(ValueError):
            self.experiment.update_tracking_metric(
                BraninMetric("m1", ["x1", "x2"]), "type2"
            )

        # Update metric definition for trial_type that doesn't exist
        with self.assertRaises(ValueError):
            self.experiment.update_tracking_metric(
                BraninMetric("m2", ["x1", "x2"]), "type3"
            )

        # Try to get runner for trial_type that's not supported
        batch = self.experiment.new_batch_trial()
        batch._trial_type = "type3"  # Force override trial type
        with self.assertRaises(ValueError):
            self.experiment.runner_for_trial(batch)

        # Try making trial with unsupported trial type
        with self.assertRaises(ValueError):
            self.experiment.new_batch_trial(trial_type="type3")

        # Try resetting runners.
        with self.assertRaises(NotImplementedError):
            self.experiment.reset_runners(SyntheticRunner())



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, Tuple

import torch
from ax.core.metric import Metric
from ax.core.search_space import SearchSpaceDigest
from ax.core.types import TCandidateMetadata
from ax.models.base import Model as BaseModel
from ax.models.types import TConfig
from botorch.acquisition.risk_measures import RiskMeasureMCObjective
from botorch.utils.datasets import SupervisedDataset
from torch import Tensor


@dataclass
class TorchOptConfig:
    """Container for lightweight representation of optimization arguments.

    This is used for communicating between modelbridge and models. This is
    an ephemeral object and not meant to be stored / serialized.

    Attributes:
        objective_weights: If doing multi-objective optimization, these denote
            which objectives should be maximized and which should be minimized.
            Otherwise, the objective is to maximize a weighted sum of
            the columns of f(x). These are the weights.
        outcome_constraints: A tuple of (A, b). For k outcome constraints
            and m outputs at f(x), A is (k x m) and b is (k x 1) such that
            A f(x) <= b.
        objective_thresholds:  A tensor containing thresholds forming a
            reference point from which to calculate pareto frontier hypervolume.
            Points that do not dominate the objective_thresholds contribute
            nothing to hypervolume.
        linear_constraints: A tuple of (A, b). For k linear constraints on
            d-dimensional x, A is (k x d) and b is (k x 1) such that
            A x <= b for feasible x.
        fixed_features: A map {feature_index: value} for features that
            should be fixed to a particular value during generation.
        pending_observations:  A list of m (k_i x d) feature tensors X
            for m outcomes and k_i pending observations for outcome i.
        model_gen_options: A config dictionary that can contain
            model-specific options. This commonly includes `optimizer_kwargs`,
            which often specifies the optimizer options to be passed to the
            optimizer while optimizing the acquisition function. These are
            generally expected to mimic the signature of `optimize_acqf`,
            though not all models may support all possible arguments and some
            models may support additional arguments that are not passed to the
            optimizer. While constructing a generation strategy, these options
            can be passed in as follows:
            >>> model_gen_kwargs = {
            >>>     "model_gen_options": {
            >>>         "optimizer_kwargs": {
            >>>             "num_restarts": 20,
            >>>             "sequential": False,
            >>>             "options": {
            >>>                 "batch_limit: 5,
            >>>                 "maxiter": 200,
            >>>             },
            >>>         },
            >>>     },
            >>> }
        rounding_func: A function that rounds an optimization result
            appropriately (i.e., according to `round-trip` transformations).
        opt_config_metrics: A dictionary of metrics that are included in
            the optimization config.
        is_moo: A boolean denoting whether this is for an MOO problem.
        risk_measure: An optional risk measure, used for robust optimization.
    """

    objective_weights: Tensor
    outcome_constraints: Optional[Tuple[Tensor, Tensor]] = None
    objective_thresholds: Optional[Tensor] = None
    linear_constraints: Optional[Tuple[Tensor, Tensor]] = None
    fixed_features: Optional[Dict[int, float]] = None
    pending_observations: Optional[List[Tensor]] = None
    model_gen_options: TConfig = field(default_factory=dict)
    rounding_func: Optional[Callable[[Tensor], Tensor]] = None
    opt_config_metrics: Optional[Dict[str, Metric]] = None
    is_moo: bool = False
    risk_measure: Optional[RiskMeasureMCObjective] = None
    fit_out_of_design: bool = False


@dataclass(frozen=True)
class TorchGenResults:
    """
    points: (n x d) Tensor of generated points.
    weights: n-tensor of weights for each point.
    gen_metadata: Generation metadata
    Dictionary of model-specific metadata for the given
                generation candidates

    """

    points: Tensor  # (n x d)-dim
    weights: Tensor  # n-dim
    gen_metadata: Dict[str, Any] = field(default_factory=dict)
    candidate_metadata: Optional[List[TCandidateMetadata]] = None


class TorchModel(BaseModel):
    """This class specifies the interface for a torch-based model.

    These methods should be implemented to have access to all of the features
    of Ax.
    """

    dtype: Optional[torch.dtype] = None
    device: Optional[torch.device] = None
    _supports_robust_optimization: bool = False

    def fit(
        self,
        datasets: List[SupervisedDataset],
        search_space_digest: SearchSpaceDigest,
        candidate_metadata: Optional[List[List[TCandidateMetadata]]] = None,
    ) -> None:
        """Fit model to m outcomes.

        Args:
            datasets: A list of ``SupervisedDataset`` containers, each
                corresponding to the data of one metric (outcome).
            search_space_digest: A ``SearchSpaceDigest`` object containing
                metadata on the features in the datasets.
            candidate_metadata: Model-produced metadata for candidates, in
                the order corresponding to the Xs.
        """
        pass

    def predict(self, X: Tensor) -> Tuple[Tensor, Tensor]:
        """Predict

        Args:
            X: (j x d) tensor of the j points at which to make predictions.

        Returns:
            2-element tuple containing

            - (j x m) tensor of outcome predictions at X.
            - (j x m x m) tensor of predictive covariances at X.
              cov[j, m1, m2] is Cov[m1@j, m2@j].
        """
        raise NotImplementedError

    def gen(
        self,
        n: int,
        search_space_digest: SearchSpaceDigest,
        torch_opt_config: TorchOptConfig,
    ) -> TorchGenResults:
        """
        Generate new candidates.

        Args:
            n: Number of candidates to generate.
            search_space_digest: A SearchSpaceDigest object containing metadata
                about the search space (e.g. bounds, parameter types).
            torch_opt_config: A TorchOptConfig object containing optimization
                arguments (e.g., objective weights, constraints).

        Returns:
            A TorchGenResult container.
        """
        raise NotImplementedError

    def best_point(
        self,
        search_space_digest: SearchSpaceDigest,
        torch_opt_config: TorchOptConfig,
    ) -> Optional[Tensor]:
        """
        Identify the current best point, satisfying the constraints in the same
        format as to gen.

        Return None if no such point can be identified.

        Args:
            search_space_digest: A SearchSpaceDigest object containing metadata
                about the search space (e.g. bounds, parameter types).
            torch_opt_config: A TorchOptConfig object containing optimization
                arguments (e.g., objective weights, constraints).

        Returns:
            d-tensor of the best point.
        """
        return None

    def cross_validate(
        self,
        datasets: List[SupervisedDataset],
        X_test: Tensor,
        search_space_digest: SearchSpaceDigest,
        use_posterior_predictive: bool = False,
    ) -> Tuple[Tensor, Tensor]:
        """Do cross validation with the given training and test sets.

        Training set is given in the same format as to fit. Test set is given
        in the same format as to predict.

        Args:
            datasets: A list of ``SupervisedDataset`` containers, each
                corresponding to the data of one metric (outcome).
            X_test: (j x d) tensor of the j points at which to make predictions.
            search_space_digest: A SearchSpaceDigest object containing
                metadata on the features in X.
            use_posterior_predictive: A boolean indicating if the predictions
                should be from the posterior predictive (i.e. including
                observation noise).

        Returns:
            2-element tuple containing

            - (j x m) tensor of outcome predictions at X.
            - (j x m x m) tensor of predictive covariances at X.
              cov[j, m1, m2] is Cov[m1@j, m2@j].
        """
        raise NotImplementedError

    def update(
        self,
        datasets: List[SupervisedDataset],
        metric_names: List[str],
        search_space_digest: SearchSpaceDigest,
        candidate_metadata: Optional[List[List[TCandidateMetadata]]] = None,
    ) -> None:
        """Update the model.

        Updating the model requires both existing and additional data.
        The data passed into this method will become the new training data.

        Args:
            datasets: A list of ``SupervisedDataset`` containers, each
                corresponding to the data of one metric (outcome). `None`
                means that there is no additional data for the corresponding
                outcome.
            metric_names: A list of metric names, with the i-th metric
                corresponding to the i-th dataset.
            search_space_digest: A SearchSpaceDigest object containing
                metadata on the features in X.
            candidate_metadata: Model-produced metadata for candidates, in
                the order corresponding to the Xs.
        """
        raise DeprecationWarning("Model.update has been deprecated. Use `fit` instead.")

    def evaluate_acquisition_function(
        self,
        X: Tensor,
        search_space_digest: SearchSpaceDigest,
        torch_opt_config: TorchOptConfig,
        acq_options: Optional[Dict[str, Any]] = None,
    ) -> Tensor:
        """Evaluate the acquisition function on the candidate set `X`.

        Args:
            X: (j x d) tensor of the j points at which to evaluate the acquisition
                function.
            search_space_digest: A dataclass used to compactly represent a search space.
            torch_opt_config: A TorchOptConfig object containing optimization
                arguments (e.g., objective weights, constraints).
            acq_options: Keyword arguments used to contruct the acquisition function.

        Returns:
            A single-element tensor with the acquisition value for these points.
        """
        raise NotImplementedError  # pragma: nocover



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import os
from typing import Any, Dict, Optional

from ax.benchmark.problems.hpo.pytorch_cnn import (
    PyTorchCNNBenchmarkProblem,
    PyTorchCNNRunner,
)
from ax.exceptions.core import UserInputError
from ax.utils.common.serialization import TClassDecoderRegistry, TDecoderRegistry
from ax.utils.common.typeutils import checked_cast
from torch.utils.data import TensorDataset

try:  # We don't require TorchVision by default.
    from torchvision import datasets, transforms

    _REGISTRY = {
        "MNIST": datasets.MNIST,
        "FashionMNIST": datasets.FashionMNIST,
    }

    if os.environ.get("TESTENV"):
        # If we are in the test environment do not download any torchvision datasets.
        # Instead, we use an empty TensorDataset
        def get_dummy_dataset(**kwargs: Dict[str, Any]) -> TensorDataset:
            return TensorDataset()

        # pyre-ignore[9] We are replacing a type with a function
        _REGISTRY = {key: get_dummy_dataset for key in _REGISTRY.keys()}


except ModuleNotFoundError:
    transforms = None
    datasets = None
    _REGISTRY = {}


class PyTorchCNNTorchvisionBenchmarkProblem(PyTorchCNNBenchmarkProblem):
    @classmethod
    def from_dataset_name(
        cls,
        name: str,
        num_trials: int,
    ) -> "PyTorchCNNTorchvisionBenchmarkProblem":
        if name not in _REGISTRY:
            raise UserInputError(
                f"Unrecognized torchvision dataset {name}. Please ensure it is listed "
                "in PyTorchCNNTorchvisionBenchmarkProblem registry."
            )
        dataset_fn = _REGISTRY[name]

        train_set = dataset_fn(
            root="./data",
            train=True,
            download=True,
            transform=transforms.ToTensor(),
        )

        test_set = dataset_fn(
            root="./data",
            train=False,
            download=True,
            transform=transforms.ToTensor(),
        )

        problem = cls.from_datasets(
            name=name,
            num_trials=num_trials,
            train_set=train_set,
            test_set=test_set,
        )
        runner = PyTorchCNNTorchvisionRunner(
            name=name, train_set=train_set, test_set=test_set
        )

        return cls(
            name=f"HPO_PyTorchCNN_Torchvision::{name}",
            search_space=problem.search_space,
            optimization_config=problem.optimization_config,
            runner=runner,
            num_trials=num_trials,
            is_noiseless=False,
            observe_noise_sd=False,
            has_ground_truth=False,
            optimal_value=problem.optimal_value,
        )


class PyTorchCNNTorchvisionRunner(PyTorchCNNRunner):
    """
    A subclass to aid in serialization. This allows us to save only the name of the
    dataset and reload it from TorchVision at deserialization time.
    """

    @classmethod
    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def serialize_init_args(cls, obj: Any) -> Dict[str, Any]:
        pytorch_cnn_runner = checked_cast(PyTorchCNNRunner, obj)

        return {"name": pytorch_cnn_runner.name}

    @classmethod
    def deserialize_init_args(
        cls,
        args: Dict[str, Any],
        decoder_registry: Optional[TDecoderRegistry] = None,
        class_decoder_registry: Optional[TClassDecoderRegistry] = None,
    ) -> Dict[str, Any]:
        name = args["name"]

        dataset_fn = _REGISTRY[name]

        train_set = dataset_fn(
            root="./data",
            train=True,
            download=True,
            transform=transforms.ToTensor(),
        )

        test_set = dataset_fn(
            root="./data",
            train=False,
            download=True,
            transform=transforms.ToTensor(),
        )

        return {"name": name, "train_set": train_set, "test_set": test_set}



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Optional

import pandas as pd
from ax.core.experiment import Experiment


class BaseAnalysis:
    """
    Abstract Analysis class for ax.
    This is an interface that defines the methods to be implemented by all analyses.
    Computes an output dataframe for each analysis
    """

    def __init__(
        self,
        experiment: Experiment,
        df_input: Optional[pd.DataFrame] = None,
        # TODO: add support for passing in experiment name, and markdown message
    ) -> None:
        """
        Initialize the analysis with the experiment object.
        For scenarios where an analysis output is already available,
        we can pass the dataframe as an input.
        """
        self._experiment = experiment
        self._df: Optional[pd.DataFrame] = df_input

    @property
    def experiment(self) -> Experiment:
        return self._experiment

    @property
    def df(self) -> pd.DataFrame:
        """
        Return the output of the analysis of this class.
        """
        if self._df is None:
            self._df = self.get_df()
        return self._df

    def get_df(self) -> pd.DataFrame:
        """
        Return the output of the analysis of this class.
        Subclasses should overwrite this.
        """
        raise NotImplementedError("get_df must be implemented by subclass")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import warnings
from collections import defaultdict
from copy import deepcopy
from logging import Logger
from typing import Callable, Dict, Iterable, List, NamedTuple, Optional, Set, Tuple
from warnings import warn

import numpy as np
from ax.core.observation import Observation, ObservationData, recombine_observations
from ax.core.optimization_config import OptimizationConfig
from ax.modelbridge.base import ModelBridge, unwrap_observation_data
from ax.utils.common.logger import get_logger
from ax.utils.stats.model_fit_stats import (
    _correlation_coefficient,
    _fisher_exact_test_p,
    _log_likelihood,
    _mape,
    _mean_prediction_ci,
    _mse,
    _rank_correlation,
    _total_raw_effect,
    _wmape,
    coefficient_of_determination,
    compute_model_fit_metrics,
    mean_of_the_standardized_error,
    ModelFitMetricProtocol,
    std_of_the_standardized_error,
)
from botorch.exceptions.warnings import InputDataWarning

logger: Logger = get_logger(__name__)

CVDiagnostics = Dict[str, Dict[str, float]]

MEAN_PREDICTION_CI = "Mean prediction CI"
MAPE = "MAPE"
wMAPE = "wMAPE"
TOTAL_RAW_EFFECT = "Total raw effect"
CORRELATION_COEFFICIENT = "Correlation coefficient"
RANK_CORRELATION = "Rank correlation"
FISHER_EXACT_TEST_P = "Fisher exact test p"
LOG_LIKELIHOOD = "Log likelihood"
MSE = "MSE"


class CVResult(NamedTuple):
    """Container for cross validation results."""

    observed: Observation
    predicted: ObservationData


class AssessModelFitResult(NamedTuple):
    """Container for model fit assessment results"""

    good_fit_metrics_to_fisher_score: Dict[str, float]
    bad_fit_metrics_to_fisher_score: Dict[str, float]


def cross_validate(
    model: ModelBridge,
    folds: int = -1,
    # pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
    test_selector: Optional[Callable] = None,
    untransform: bool = True,
    use_posterior_predictive: bool = False,
) -> List[CVResult]:
    """Cross validation for model predictions.

    Splits the model's training data into train/test folds and makes
    out-of-sample predictions on the test folds.

    Train/test splits are made based on arm names, so that repeated
    observations of a arm will always be in the train or test set
    together.

    The test set can be limited to a specific set of observations by passing in
    a test_selector callable. This function should take in an Observation
    and return a boolean indiciating if it should be used in the test set or
    not. For example, we can limit the test set to arms with trial 0 with
    test_selector = lambda obs: obs.features.trial_index == 0
    If not provided, all observations will be available for the test set.

    Args:
        model: Fitted model (ModelBridge) to cross validate.
        folds: Number of folds. Use -1 for leave-one-out, otherwise will be
            k-fold.
        test_selector: Function for selecting observations for the test set.
        untransform: Whether to untransform the model predictions before
            cross validating.
            Models are trained on transformed data, and candidate generation
            is performed in the transformed space. Computing the model
            quality metric based on the cross-validation results in the
            untransformed space may not be representative of the model that
            is actually used for candidate generation in case of non-invertible
            transforms, e.g., Winsorize or LogY.
            While the model in the transformed space may not be representative
            of the original data in regions where outliers have been removed,
            we have found it to better reflect the how good the model used
            for candidate generation actually is.
        use_posterior_predictive: A boolean indicating if the predictions
            should be from the posterior predictive (i.e. including
            observation noise). Note: we should reconsider how we compute
            cross-validation and model fit metrics where there is non-
            Gaussian noise.

    Returns:
        A CVResult for each observation in the training data.
    """
    # Get in-design training points
    training_data = [
        obs
        for i, obs in enumerate(model.get_training_data())
        if model.training_in_design[i]
    ]
    arm_names = {obs.arm_name for obs in training_data}
    n = len(arm_names)
    if folds > n:
        raise ValueError(f"Training data only has {n} arms, which is less than folds")
    elif n == 0:
        raise ValueError(
            f"{model.__class__.__name__} has no training data.  Either it has been "
            "incorrectly initialized or should not be cross validated."
        )
    elif folds < 2 and folds != -1:
        raise ValueError("Folds must be -1 for LOO, or > 1.")
    elif folds == -1:
        folds = n

    arm_names_rnd = np.array(list(arm_names))
    # Not necessary to shuffle when using LOO, avoids differences in floating point
    # computations making equality tests brittle.
    if folds != -1:
        np.random.shuffle(arm_names_rnd)
    result = []
    for train_names, test_names in _gen_train_test_split(
        folds=folds, arm_names=arm_names_rnd
    ):
        # Construct train/test data
        cv_training_data = []
        cv_test_data = []
        cv_test_points = []
        for obs in training_data:
            if obs.arm_name in train_names:
                cv_training_data.append(obs)
            elif obs.arm_name in test_names and (
                test_selector is None or test_selector(obs)
            ):
                cv_test_points.append(obs.features)
                cv_test_data.append(obs)
        if len(cv_test_points) == 0:
            continue

        # Make the prediction
        if untransform:
            cv_test_predictions = model.cross_validate(
                cv_training_data=cv_training_data,
                cv_test_points=cv_test_points,
                use_posterior_predictive=use_posterior_predictive,
            )
        else:
            # Get test predictions in transformed space
            (
                cv_training_data,
                cv_test_points,
                search_space,
            ) = model._transform_inputs_for_cv(
                cv_training_data=cv_training_data, cv_test_points=cv_test_points
            )
            with warnings.catch_warnings():
                # Since each CV fold removes points from the training data, the
                # remaining observations will not pass the standardization test.
                # To avoid confusing users with this warning, we filter it out.
                warnings.filterwarnings(
                    "ignore",
                    message="Data is not standardized",
                    category=InputDataWarning,
                )
                cv_test_predictions = model._cross_validate(
                    search_space=search_space,
                    cv_training_data=cv_training_data,
                    cv_test_points=cv_test_points,
                    use_posterior_predictive=use_posterior_predictive,
                )
            # Get test observations in transformed space
            cv_test_data = deepcopy(cv_test_data)
            for t in model.transforms.values():
                cv_test_data = t.transform_observations(cv_test_data)
        # Form CVResult objects
        for i, obs in enumerate(cv_test_data):
            result.append(CVResult(observed=obs, predicted=cv_test_predictions[i]))
    return result


def cross_validate_by_trial(
    model: ModelBridge, trial: int = -1, use_posterior_predictive: bool = False
) -> List[CVResult]:
    """Cross validation for model predictions on a particular trial.

    Uses all of the data up until the specified trial to predict each of the
    arms that was launched in that trial. Defaults to the last trial.

    Args:
        model: Fitted model (ModelBridge) to cross validate.
        trial: Trial for which predictions are evaluated.
        use_posterior_predictive: A boolean indicating if the predictions
            should be from the posterior predictive (i.e. including
            observation noise).

    Returns:
        A CVResult for each observation in the training data.
    """
    # Get in-design training points
    training_data = [
        obs
        for i, obs in enumerate(model.get_training_data())
        if model.training_in_design[i]
    ]
    all_trials = {
        int(d.features.trial_index)
        for d in training_data
        if d.features.trial_index is not None
    }
    if len(all_trials) < 2:
        raise ValueError(f"Training data has fewer than 2 trials ({all_trials})")
    if trial < 0:
        trial = max(all_trials)
    elif trial not in all_trials:
        raise ValueError(f"Trial {trial} not found in training data")
    # Construct train/test data
    cv_training_data = []
    cv_test_data = []
    cv_test_points = []
    for obs in training_data:
        if obs.features.trial_index is None:
            continue
        elif obs.features.trial_index < trial:
            cv_training_data.append(obs)
        elif obs.features.trial_index == trial:
            cv_test_points.append(obs.features)
            cv_test_data.append(obs)
    # Make the prediction
    cv_test_predictions = model.cross_validate(
        cv_training_data=cv_training_data,
        cv_test_points=cv_test_points,
        use_posterior_predictive=use_posterior_predictive,
    )
    # Form CVResult objects
    result = [
        CVResult(observed=obs, predicted=cv_test_predictions[i])
        for i, obs in enumerate(cv_test_data)
    ]
    return result


def compute_diagnostics(result: List[CVResult]) -> CVDiagnostics:
    """Computes diagnostics for given cross validation results.

    It provides a dictionary with values for the following diagnostics, for
    each metric:

    - 'Mean prediction CI': the average width of the CIs at each of the CV
      predictions, relative to the observed mean.
    - 'MAPE': mean absolute percentage error of the estimated mean relative
      to the observed mean.
    - 'wMAPE': Weighted mean absolute percentage error.
    - 'Total raw effect': the multiple change from the smallest observed
      mean to the largest observed mean, i.e. `(max - min) / min`.
    - 'Correlation coefficient': the Pearson correlation of the estimated
      and observed means.
    - 'Rank correlation': the Spearman correlation of the estimated
      and observed means.
    - 'Fisher exact test p': we test if the model is able to distinguish the
      bottom half of the observations from the top half, using Fisher's
      exact test and the observed/estimated means. A low p value indicates
      that the model has some ability to identify good arms. A high p value
      indicates that the model cannot identify arms better than chance, or
      that the observations are too noisy to be able to tell.

    Each of these is returned as a dictionary from metric name to value for
    that metric.

    Args:
        result: Output of cross_validate

    Returns:
        A dictionary keyed by diagnostic name with results as described above.
    """
    # Extract per-metric outcomes from CVResults.
    y_obs = defaultdict(list)
    y_pred = defaultdict(list)
    se_pred = defaultdict(list)
    for res in result:
        for j, metric_name in enumerate(res.observed.data.metric_names):
            y_obs[metric_name].append(res.observed.data.means[j])
            # Find the matching prediction
            k = res.predicted.metric_names.index(metric_name)
            y_pred[metric_name].append(res.predicted.means[k])
            se_pred[metric_name].append(np.sqrt(res.predicted.covariance[k, k]))
    y_obs = _arrayify_dict_values(y_obs)
    y_pred = _arrayify_dict_values(y_pred)
    se_pred = _arrayify_dict_values(se_pred)

    diagnostic_fns = {
        MEAN_PREDICTION_CI: _mean_prediction_ci,
        MAPE: _mape,
        wMAPE: _wmape,
        TOTAL_RAW_EFFECT: _total_raw_effect,
        CORRELATION_COEFFICIENT: _correlation_coefficient,
        RANK_CORRELATION: _rank_correlation,
        FISHER_EXACT_TEST_P: _fisher_exact_test_p,
        LOG_LIKELIHOOD: _log_likelihood,
        MSE: _mse,
    }
    diagnostics = compute_model_fit_metrics(
        y_obs=y_obs, y_pred=y_pred, se_pred=se_pred, fit_metrics_dict=diagnostic_fns
    )
    return diagnostics


def _arrayify_dict_values(d: Dict[str, List[float]]) -> Dict[str, np.ndarray]:
    """Helper to convert dictionary values to numpy arrays."""
    return {k: np.array(v) for k, v in d.items()}


def assess_model_fit(
    diagnostics: CVDiagnostics,
    significance_level: float = 0.1,
) -> AssessModelFitResult:
    """Assess model fit for given diagnostics results.

    It determines if a model fit is good or bad based on Fisher exact test p

    Args:
        diagnostics: Output of compute_diagnostics

    Returns:
        Two dictionaries, one for good metrics, one for bad metrics, each
        mapping metric name to p-value
    """

    good_fit_metrics_to_fisher_score: Dict[str, float] = {}
    bad_fit_metrics_to_fisher_score: Dict[str, float] = {}

    for metric, score in diagnostics[FISHER_EXACT_TEST_P].items():
        if score > significance_level:
            bad_fit_metrics_to_fisher_score[metric] = score
        else:
            good_fit_metrics_to_fisher_score[metric] = score
    if len(bad_fit_metrics_to_fisher_score) > 0:
        logger.warning(
            "{} {} {} unable to be reliably fit.".format(
                ("Metrics" if len(bad_fit_metrics_to_fisher_score) > 1 else "Metric"),
                (" , ".join(bad_fit_metrics_to_fisher_score.keys())),
                ("were" if len(bad_fit_metrics_to_fisher_score) > 1 else "was"),
            )
        )
    return AssessModelFitResult(
        good_fit_metrics_to_fisher_score=good_fit_metrics_to_fisher_score,
        bad_fit_metrics_to_fisher_score=bad_fit_metrics_to_fisher_score,
    )


def has_good_opt_config_model_fit(
    optimization_config: OptimizationConfig,
    assess_model_fit_result: AssessModelFitResult,
) -> bool:
    """Assess model fit for given diagnostics results across the optimization
    config metrics

    Bad fit criteria: Any objective metrics are poorly fit based on
    the Fisher exact test p (see assess_model_fit())

    TODO[]: Incl. outcome constraints in assessment

    Args:
        optimization_config: Objective/Outcome constraint metrics to assess
        diagnostics: Output of compute_diagnostics

    Returns:
        Two dictionaries, one for good metrics, one for bad metrics, each
        mapping metric name to p-value
    """

    # Bad fit criteria: Any objective metrics are poorly fit
    # TODO[]: Incl. outcome constraints in assessment
    has_good_opt_config_fit = all(
        (m.name in assess_model_fit_result.good_fit_metrics_to_fisher_score)
        for m in optimization_config.objective.metrics
    )
    return has_good_opt_config_fit


def _gen_train_test_split(
    folds: int, arm_names: np.ndarray
) -> Iterable[Tuple[Set[str], Set[str]]]:
    """Return train/test splits of arm names.

    Args:
        folds: Number of folds to return
        arm_names: Array of arm names

    Returns:
        Yields (train, test) tuple of arm names.
    """
    n = len(arm_names)
    test_size = n // folds  # The size of all test sets but the last
    final_size = test_size + (n - folds * test_size)  # Grab the leftovers
    for fold in range(folds):
        # We will take the test set from the back of the array.
        # Roll the list of arm names to get a fresh test set
        arm_names = np.roll(arm_names, test_size)
        n_test = test_size if fold < folds - 1 else final_size
        yield set(arm_names[:-n_test]), set(arm_names[-n_test:])


"""
############################## Model Fit Metrics Utils ##############################
"""


def get_fit_and_std_quality_and_generalization_dict(
    fitted_model_bridge: ModelBridge,
) -> Dict[str, Optional[float]]:
    """
    Get stats and gen from a fitted ModelBridge for analytics purposes.
    """
    try:
        model_fit_dict = compute_model_fit_metrics_from_modelbridge(
            model_bridge=fitted_model_bridge,
            generalization=False,
            untransform=False,
        )
        # similar for uncertainty quantification, but distance from 1 matters
        std = list(model_fit_dict["std_of_the_standardized_error"].values())

        # generalization metrics
        model_gen_dict = compute_model_fit_metrics_from_modelbridge(
            model_bridge=fitted_model_bridge,
            generalization=True,
            untransform=False,
        )
        gen_std = list(model_gen_dict["std_of_the_standardized_error"].values())
        return {
            "model_fit_quality": _model_fit_metric(model_fit_dict),
            "model_std_quality": _model_std_quality(np.array(std)),
            "model_fit_generalization": _model_fit_metric(model_gen_dict),
            "model_std_generalization": _model_std_quality(np.array(gen_std)),
        }

    except Exception as e:
        warn("Encountered exception in computing model fit quality: " + str(e))
        return {
            "model_fit_quality": None,
            "model_std_quality": None,
            "model_fit_generalization": None,
            "model_std_generalization": None,
        }


def compute_model_fit_metrics_from_modelbridge(
    model_bridge: ModelBridge,
    fit_metrics_dict: Optional[Dict[str, ModelFitMetricProtocol]] = None,
    generalization: bool = False,
    untransform: bool = False,
) -> Dict[str, Dict[str, float]]:
    """Computes the model fit metrics given a ModelBridge and an Experiment.

    Args:
        model_bridge: The ModelBridge for which to compute the model fit metrics.
        experiment: The experiment with whose data to compute the metrics if
            generalization == False. Otherwise, the data is taken from the ModelBridge.
        fit_metrics_dict: An optional dictionary with model fit metric functions,
            i.e. a ModelFitMetricProtocol, as values and their names as keys.
        generalization: Boolean indicating whether to compute the generalization
            metrics on cross-validation data or on the training data. The latter
            helps diagnose problems with model training, rather than generalization.
        untransform: Boolean indicating whether to untransform model predictions
            before calcualting the model fit metrics. False by default as models
            are trained in transformed space and model fit should be
            evaluated in transformed space.

    Returns:
        A nested dictionary mapping from the *model fit* metric names and the
        *experimental metric* names to the values of the model fit metrics.

        Example for an imaginary AutoML experiment that seeks to minimize the test
        error after training an expensive model, with respect to hyper-parameters:

        ```
        model_fit_dict = compute_model_fit_metrics_from_modelbridge(model_bridge, exp)
        model_fit_dict["coefficient_of_determination"]["test error"] =
            `coefficient of determination of the test error predictions`
        ```
    """
    predict_func = (
        _predict_on_cross_validation_data
        if generalization
        else _predict_on_training_data
    )
    y_obs, y_pred, se_pred = predict_func(
        model_bridge=model_bridge, untransform=untransform
    )
    if fit_metrics_dict is None:
        fit_metrics_dict = {
            "coefficient_of_determination": coefficient_of_determination,
            "mean_of_the_standardized_error": mean_of_the_standardized_error,
            "std_of_the_standardized_error": std_of_the_standardized_error,
        }

    return compute_model_fit_metrics(
        y_obs=y_obs,
        y_pred=y_pred,
        se_pred=se_pred,
        fit_metrics_dict=fit_metrics_dict,
    )


def _model_fit_metric(metric_dict: Dict[str, Dict[str, float]]) -> float:
    # We'd ideally log the entire `model_fit_dict` as a single model fit metric
    # can't capture the nuances of multiple experimental metrics, but this might
    # lead to database performance issues. So instead, we take the worst
    # coefficient of determination as model fit quality and store the full data
    # in Manifold (TODO).
    return min(metric_dict["coefficient_of_determination"].values())


def _model_std_quality(std: np.ndarray) -> float:
    """Quantifies quality of the model uncertainty. A value of one means the
    uncertainty is perfectly predictive of the true standard deviation of the error.
    Values larger than one indicate over-estimation and negative values indicate
    under-estimation of the true standard deviation of the error. In particular, a value
    of 2 (resp. 1 / 2) represents an over-estimation (resp. under-estimation) of the
    true standard deviation of the error by a factor of 2.

    Args:
        std: The standard deviation of the standardized error.

    Returns:
        The factor corresponding to the worst over- or under-estimation factor of the
        standard deviation of the error among all experimentally observed metrics.
    """
    max_std, min_std = np.max(std), np.min(std)
    # comparing worst over-estimation factor with worst under-estimation factor
    inv_model_std_quality = max_std if max_std > 1 / min_std else min_std
    # reciprocal so that values greater than one indicate over-estimation and
    # values smaller than indicate underestimation of the uncertainty.
    return 1 / inv_model_std_quality


def _predict_on_training_data(
    model_bridge: ModelBridge,
    untransform: bool = False,
) -> Tuple[
    Dict[str, np.ndarray],
    Dict[str, np.ndarray],
    Dict[str, np.ndarray],
]:
    """Makes predictions on the training data of a given experiment using a ModelBridge
    and returning the observed values, and the corresponding predictive means and
    predictive standard deviations of the model, in transformed space.

    NOTE: This is a helper function for `compute_model_fit_metrics_from_modelbridge`.

    Args:
        model_bridge: A ModelBridge object with which to make predictions.
        untransform: Boolean indicating whether to untransform model predictions.

    Returns:
        A tuple containing three dictionaries for 1) observed metric values, and the
        model's associated 2) predictive means and 3) predictive standard deviations.
    """
    observations = model_bridge.get_training_data()  # List[Observation]

    # NOTE: the following up to the end of the untransform block could be replaced
    # with model_bridge's public predict / private _batch_predict method, if the
    # latter had a boolean untransform flag.

    # Transform observations -- this will transform both obs data and features
    for t in model_bridge.transforms.values():
        observations = t.transform_observations(observations)

    observation_features = [obs.features for obs in observations]

    # Make predictions in transformed space
    observation_data_pred = model_bridge._predict(observation_features)

    if untransform:
        # Apply reverse transforms, in reverse order
        pred_observations = recombine_observations(
            observation_features=observation_features,
            observation_data=observation_data_pred,
        )
        for t in reversed(list(model_bridge.transforms.values())):
            pred_observations = t.untransform_observations(pred_observations)

        observation_data_pred = [obs.data for obs in pred_observations]

    mean_predicted, cov_predicted = unwrap_observation_data(observation_data_pred)
    mean_observed = [
        obs.data.means_dict for obs in observations
    ]  # List[Dict[str, float]]

    metric_names = observations[0].data.metric_names
    mean_observed = _list_of_dicts_to_dict_of_lists(
        list_of_dicts=mean_observed, keys=metric_names
    )
    # converting dictionary values to arrays
    mean_observed = {k: np.array(v) for k, v in mean_observed.items()}
    mean_predicted = {k: np.array(v) for k, v in mean_predicted.items()}
    std_predicted = {m: np.sqrt(np.array(cov_predicted[m][m])) for m in cov_predicted}
    return mean_observed, mean_predicted, std_predicted


def _predict_on_cross_validation_data(
    model_bridge: ModelBridge,
    untransform: bool = False,
) -> Tuple[
    Dict[str, np.ndarray],
    Dict[str, np.ndarray],
    Dict[str, np.ndarray],
]:
    """Makes leave-one-out cross-validation predictions on the training data of the
    ModelBridge and returns the observed values, and the corresponding predictive means
    and predictive standard deviations of the model as numpy arrays,
    in transformed space.

    NOTE: This is a helper function for `compute_model_fit_metrics_from_modelbridge`.

    Args:
        model_bridge: A ModelBridge object with which to make predictions.
        untransform: Boolean indicating whether to untransform model predictions
            before cross validating. False by default.

    Returns:
        A tuple containing three dictionaries, each mapping metric_name to:
            1. observed metric values,
            2. LOOCV predicted mean at each observed point, and
            3. LOOCV predicted standard deviation at each observed point.
    """
    # IDEA: could use cross_validate_by_trial on the last few trials, since the
    # cross-validation performance on these is likely more correlated with the
    # performance of the model on an upcoming trial due to the sequential nature
    # of the data generated by BO.
    cv = cross_validate(model=model_bridge, untransform=untransform)

    metric_names = cv[0].observed.data.metric_names
    mean_observed = {k: [] for k in metric_names}
    mean_predicted = {k: [] for k in metric_names}
    std_predicted = {k: [] for k in metric_names}

    for cvi in cv:
        obs = cvi.observed.data
        for k, v in zip(obs.metric_names, obs.means):
            mean_observed[k].append(v)

        pred = cvi.predicted
        for k, v in zip(pred.metric_names, pred.means):
            mean_predicted[k].append(v)

        pred_se = np.sqrt(pred.covariance.diagonal().clip(0))
        for k, v in zip(pred.metric_names, pred_se):
            std_predicted[k].append(v)

    mean_observed = {k: np.array(v) for k, v in mean_observed.items()}
    mean_predicted = {k: np.array(v) for k, v in mean_predicted.items()}
    std_predicted = {k: np.array(v) for k, v in std_predicted.items()}
    return mean_observed, mean_predicted, std_predicted


def _list_of_dicts_to_dict_of_lists(
    list_of_dicts: List[Dict[str, float]], keys: List[str]
) -> Dict[str, List[float]]:
    """Converts a list of dicts indexed by a string to a dict of lists."""
    return {key: [d[key] for d in list_of_dicts] for key in keys}



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from typing import Any, Dict, List, Optional

import pandas as pd

from ax.analysis.base_plotly_visualization import BasePlotlyVisualization

from ax.analysis.helpers.cross_validation_helpers import (
    cv_results_to_df,
    diagonal_trace,
    get_plotting_limit_ignore_outliers,
)

from ax.analysis.helpers.layout_helpers import layout_format, updatemenus_format

from ax.analysis.helpers.scatter_helpers import (
    error_scatter_trace_from_df,
    extract_mean_and_error_from_df,
)

from ax.core.experiment import Experiment
from ax.modelbridge import ModelBridge

from ax.modelbridge.cross_validation import cross_validate, CVResult

from plotly import graph_objs as go


class CrossValidationPlot(BasePlotlyVisualization):
    CROSS_VALIDATION_CAPTION = (
        "<b>NOTE:</b> We have tried our best to only plot the region of interest.<br>"
        "This may hide outliers. You can autoscale the axes to see all trials."
    )

    def __init__(
        self,
        experiment: Experiment,
        model: ModelBridge,
        label_dict: Optional[Dict[str, str]] = None,
        caption: str = CROSS_VALIDATION_CAPTION,
    ) -> None:
        """
        Args:
        experiment: Experiment containing trials to plot
        model: ModelBridge to cross validate against
        label_dict: optional map from real metric names to shortened names
        caption: text to display below the plot
        """
        self.model = model
        self.cv: List[CVResult] = cross_validate(model=model)

        self.label_dict: Optional[Dict[str, str]] = label_dict
        if self.label_dict:
            self.cv = self.remap_label(cv_results=self.cv, label_dict=self.label_dict)

        self.metric_names: List[str] = list(
            set().union(*(cv_result.predicted.metric_names for cv_result in self.cv))
        )
        self.caption = caption

        super().__init__(experiment=experiment)

    def get_df(self) -> pd.DataFrame:
        """
        Overrides BaseAnalysis.get_df()

        Returns:
            df representation of the cross validation results.
            columns:
            {
                "arm_name": name of the arm in the cross validation result
                "metric_name": name of the observed/predicted metric
                "x": value observed for the metric for this arm
                "x_se": standard error of observed metric (0 for observations)
                "y": value predicted for the metric for this arm
                "y_se": standard error of predicted metric for this arm
                "arm_parameters": Parametrization of the arm
            }
        """

        df = pd.concat(
            [
                cv_results_to_df(
                    cv_results=self.cv,
                    metric_name=metric,
                )
                for metric in self.metric_names
            ]
        )

        return df

    @staticmethod
    def compose_annotation(
        caption: str, x: float = 0.0, y: float = -0.15
    ) -> List[Dict[str, Any]]:
        """Composes an annotation dict for use in Plotly figure.
        args:
            caption: str to use for dropdown text
            x: x position of the annotation
            y: y position of the annotation

        returns:
            Annotation dict for use in Plotly figure.
        """
        return [
            {
                "showarrow": False,
                "text": caption,
                "x": x,
                "xanchor": "left",
                "xref": "paper",
                "y": y,
                "yanchor": "top",
                "yref": "paper",
                "align": "left",
            },
        ]

    @staticmethod
    def remap_label(
        cv_results: List[CVResult], label_dict: Dict[str, str]
    ) -> List[CVResult]:
        """Remaps labels in cv_results according to label_dict.

        Args:
            cv_results: A CVResult for each observation in the training data.
            label_dict: optional map from real metric names to shortened names

        Returns:
            A CVResult with metric names mapped from label_dict.
        """
        cv_results = deepcopy(cv_results)  # Copy and edit in-place
        for cv_i in cv_results:
            cv_i.observed.data.metric_names = [
                label_dict.get(m, m) for m in cv_i.observed.data.metric_names
            ]
            cv_i.predicted.metric_names = [
                label_dict.get(m, m) for m in cv_i.predicted.metric_names
            ]
        return cv_results

    def obs_vs_pred_dropdown_plot(
        self,
        xlabel: str = "Actual Outcome",
        ylabel: str = "Predicted Outcome",
    ) -> go.Figure:
        """Plot a dropdown plot of observed vs. predicted values from the
        cross validation results.

        Args:
            xlabel: Label for x-axis.
            ylabel: Label for y-axis.
        """
        traces = []
        metric_dropdown = []
        layout_axis_range = []

        # Get the union of all metric_names seen in predictions
        metric_names = self.metric_names
        df = self.get_df()

        for i, metric in enumerate(metric_names):
            metric_filtered_df = df.loc[df["metric_name"] == metric]

            y_raw, se_raw, y_hat, se_hat = extract_mean_and_error_from_df(
                metric_filtered_df
            )

            # Use the min/max of the limits
            layout_range, diagonal_trace_range = get_plotting_limit_ignore_outliers(
                x=y_raw, y=y_hat, se_x=se_raw, se_y=se_hat
            )
            layout_axis_range.append(layout_range)

            # add a diagonal dotted line to plot
            traces.append(
                diagonal_trace(
                    diagonal_trace_range[0],
                    diagonal_trace_range[1],
                    visible=(i == 0),
                )
            )

            traces.append(
                error_scatter_trace_from_df(
                    df=metric_filtered_df,
                    show_CI=True,
                    visible=(i == 0),
                    x_axis_label="Actual Outcome",
                    y_axis_label="Predicted Outcome",
                )
            )

            # only the first two traces are visible (corresponding to first outcome
            # in dropdown)
            is_visible = [False] * (len(metric_names) * 2)
            is_visible[2 * i] = True
            is_visible[2 * i + 1] = True

            # on dropdown change, restyle
            metric_dropdown.append(
                {
                    "args": [
                        {"visible": is_visible},
                        {
                            "xaxis.range": layout_axis_range[-1],
                            "yaxis.range": layout_axis_range[-1],
                        },
                    ],
                    "label": metric,
                    "method": "update",
                }
            )

        updatemenus = updatemenus_format(metric_dropdown=metric_dropdown)
        layout = layout_format(
            layout_axis_range_value=layout_axis_range[0],
            xlabel=xlabel,
            ylabel=ylabel,
            updatemenus=updatemenus,
        )

        return go.Figure(data=traces, layout=layout)

    def get_fig(self) -> go.Figure:
        """
        Interactive cross-validation (CV) plotting; select metric via dropdown.
        Note: uses the Plotly version of dropdown (which means that all data is
        stored within the notebook).

        Returns:
            go.Figure: Plotly figure with cross validation plot
        """
        caption = self.caption

        fig = self.obs_vs_pred_dropdown_plot()

        current_bmargin = fig["layout"]["margin"].b or 90
        caption_height = 100 * (len(caption) > 0)
        fig["layout"]["margin"].b = current_bmargin + caption_height
        fig["layout"]["height"] += caption_height
        fig["layout"]["annotations"] += tuple(self.compose_annotation(caption))
        fig["layout"]["title"] = "Cross-validation"
        return fig



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Optional

import pandas as pd

import plotly.graph_objects as go

from ax.analysis.base_analysis import BaseAnalysis
from ax.core.experiment import Experiment


class BasePlotlyVisualization(BaseAnalysis):
    """
    Abstract PlotlyVisualization class for ax.
    This is an interface that defines the method to be implemented by all ax plots.
    Computes an output dataframe for each analysis
    """

    def __init__(
        self,
        experiment: Experiment,
        df_input: Optional[pd.DataFrame] = None,
        fig_input: Optional[go.Figure] = None,
    ) -> None:
        """
        Initialize the analysis with the experiment object.
        For scenarios where an analysis output is already available,
        we can pass the dataframe as an input.
        """
        self._fig = fig_input
        super().__init__(experiment=experiment, df_input=df_input)

    @property
    def fig(self) -> go.Figure:
        """
        Return the output of the analysis of this class.
        """
        if self._fig is None:
            self._fig = self.get_fig()
        return self._fig

    def get_fig(self) -> go.Figure:
        """
        Return the plotly figure of the analysis of this class.
        Subclasses should overwrite this.
        """
        raise NotImplementedError("get_fig must be implemented by subclass")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from logging import Logger
from typing import Any, Callable, Dict, List, Optional, Sequence, Set, Tuple, Union

# Module-level import to avoid circular dependency b/w this file and
# generation_strategy.py
from ax import modelbridge
from ax.core.arm import Arm
from ax.core.base_trial import TrialStatus
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.observation import ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.core.search_space import SearchSpace
from ax.exceptions.core import UserInputError
from ax.exceptions.generation_strategy import GenerationStrategyRepeatedPoints
from ax.modelbridge.base import ModelBridge
from ax.modelbridge.best_model_selector import BestModelSelector
from ax.modelbridge.model_spec import FactoryFunctionModelSpec, ModelSpec
from ax.modelbridge.registry import ModelRegistryBase
from ax.modelbridge.transition_criterion import (
    MaxGenerationParallelism,
    MaxTrials,
    MinTrials,
    TransitionCriterion,
    TrialBasedCriterion,
)
from ax.utils.common.base import Base, SortableBase
from ax.utils.common.logger import get_logger
from ax.utils.common.serialization import SerializationMixin
from ax.utils.common.typeutils import not_none


logger: Logger = get_logger(__name__)

TModelFactory = Callable[..., ModelBridge]
CANNOT_SELECT_ONE_MODEL_MSG = """\
Base `GenerationNode` does not implement selection among fitted \
models, so exactly one `ModelSpec` must be specified when using \
`GenerationNode._pick_fitted_model_to_gen_from` (usually called \
by `GenerationNode.gen`.
"""
MAX_GEN_DRAWS = 5
MAX_GEN_DRAWS_EXCEEDED_MESSAGE = (
    f"GenerationStrategy exceeded `MAX_GEN_DRAWS` of {MAX_GEN_DRAWS} while trying to "
    "generate a unique parameterization. This indicates that the search space has "
    "likely been fully explored, or that the sweep has converged."
)


class GenerationNode(SerializationMixin, SortableBase):
    """Base class for GenerationNode, capable of fitting one or more model specs under
    the hood and generating candidates from them.

    Args:
        node_name: A unique name for the GenerationNode. Used for storage purposes.
        model_specs: A list of ModelSpecs to be selected from for generation in this
            GenerationNode.
        best_model_selector: A ``BestModelSelector`` used to select the ``ModelSpec``
            to generate from in ``GenerationNode`` with multiple ``ModelSpec``s.
        should_deduplicate: Whether to deduplicate the parameters of proposed arms
            against those of previous arms via rejection sampling. If this is True,
            the GenerationStrategy will discard generator runs produced from the
            GenerationNode that has `should_deduplicate=True` if they contain arms
            already present on the experiment and replace them with new generator runs.
            If no generator run with entirely unique arms could be produced in 5
            attempts, a `GenerationStrategyRepeatedPoints` error will be raised, as we
            assume that the optimization converged when the model can no longer suggest
            unique arms.
        transition_criteria: List of TransitionCriterion, each of which describes a
            condition that must be met before completing a GenerationNode. All `is_met`
            must evaluateTrue for the GenerationStrategy to move on to the next
            GenerationNode.

    Note for developers: by "model" here we really mean an Ax ModelBridge object, which
    contains an Ax Model under the hood. We call it "model" here to simplify and focus
    on explaining the logic of GenerationStep and GenerationStrategy.
    """

    # Required options:
    model_specs: List[ModelSpec]
    # TODO: Move `should_deduplicate` to `ModelSpec` if possible, and make optional
    should_deduplicate: bool
    _node_name: str

    # Optional specifications
    _model_spec_to_gen_from: Optional[ModelSpec] = None
    # TODO: @mgarrard should this be a dict criterion_class name -> criterion mapping?
    _transition_criteria: Optional[Sequence[TransitionCriterion]]

    # [TODO] Handle experiment passing more eloquently by enforcing experiment
    # attribute is set in generation strategies class
    _generation_strategy: Optional[
        modelbridge.generation_strategy.GenerationStrategy
    ] = None

    def __init__(
        self,
        node_name: str,
        model_specs: List[ModelSpec],
        best_model_selector: Optional[BestModelSelector] = None,
        should_deduplicate: bool = False,
        transition_criteria: Optional[Sequence[TransitionCriterion]] = None,
    ) -> None:
        self._node_name = node_name
        # While `GenerationNode` only handles a single `ModelSpec` in the `gen`
        # and `_pick_fitted_model_to_gen_from` methods, we validate the
        # length of `model_specs` in `_pick_fitted_model_to_gen_from` in order
        # to not require all `GenerationNode` subclasses to override an `__init__`
        # method to bypass that validation.
        self.model_specs = model_specs
        self.best_model_selector = best_model_selector
        self.should_deduplicate = should_deduplicate
        self._transition_criteria = transition_criteria

    @property
    def node_name(self) -> str:
        """Returns the unique name of this GenerationNode"""
        return self._node_name

    @property
    def model_spec_to_gen_from(self) -> ModelSpec:
        """Returns the cached `_model_spec_to_gen_from` or gets it from
        `_pick_fitted_model_to_gen_from` and then caches and returns it
        """
        if self._model_spec_to_gen_from is not None:
            return self._model_spec_to_gen_from

        self._model_spec_to_gen_from = self._pick_fitted_model_to_gen_from()
        return self._model_spec_to_gen_from

    @property
    def model_to_gen_from_name(self) -> Optional[str]:
        """Returns the name of the model that will be used for gen, if there is one.
        Otherwise, returns None.
        """
        if self._model_spec_to_gen_from is not None:
            return self._model_spec_to_gen_from.model_key
        else:
            return None

    @property
    def generation_strategy(self) -> modelbridge.generation_strategy.GenerationStrategy:
        """Returns a backpointer to the GenerationStrategy, useful for obtaining the
        experiment associated with this GenerationStrategy"""
        # TODO: @mgarrard remove this property once we make experiment a required
        # argument on GenerationStrategy
        if self._generation_strategy is None:
            raise ValueError(
                "Generation strategy has not been initialized on this node."
            )
        return not_none(self._generation_strategy)

    @property
    def transition_criteria(self) -> Sequence[TransitionCriterion]:
        """Returns the sequence of TransitionCriteria that will be used to determine
        if this GenerationNode is complete and should transition to the next node.
        """
        return [] if self._transition_criteria is None else self._transition_criteria

    @property
    def experiment(self) -> Experiment:
        """Returns the experiment associated with this GenerationStrategy"""
        return self.generation_strategy.experiment

    @property
    def is_completed(self) -> bool:
        """Returns True if this GenerationNode is complete and should transition to
        the next node.
        """
        return self.should_transition_to_next_node(raise_data_required_error=False)[0]

    @property
    def _unique_id(self) -> str:
        """Returns a unique (w.r.t. parent class: ``GenerationStrategy``) id
        for this GenerationNode. Used for SQL storage.
        """
        return self.node_name

    @property
    def _fitted_model(self) -> Optional[ModelBridge]:
        """Private property to return optional fitted_model from
        self.model_spec_to_gen_from for convenience. If no model is fit,
        will return None. If using the non-private `fitted_model` property,
        and no model is fit, a UserInput error will be raised.
        """
        return self.model_spec_to_gen_from._fitted_model

    def fit(
        self,
        experiment: Experiment,
        data: Data,
        search_space: Optional[SearchSpace] = None,
        optimization_config: Optional[OptimizationConfig] = None,
        **kwargs: Any,
    ) -> None:
        """Fits the specified models to the given experiment + data using
        the model kwargs set on each corresponding model spec and the kwargs
        passed to this method.

        Args:
            experiment: The experiment to fit the model to.
            data: The experiment data used to fit the model.
            search_space: An optional overwrite for the experiment search space.
            optimization_config: An optional overwrite for the experiment
                optimization config.
            kwargs: Additional keyword arguments to pass to the model's
                ``fit`` method. NOTE: Local kwargs take precedence over the ones
                stored in ``ModelSpec.model_kwargs``.
        """
        self._model_spec_to_gen_from = None
        for model_spec in self.model_specs:
            model_spec.fit(  # Stores the fitted model as `model_spec._fitted_model`
                experiment=experiment,
                data=data,
                search_space=search_space,
                optimization_config=optimization_config,
                **kwargs,
            )

    # TODO [drfreund]: Move this up to `GenerationNodeInterface` once implemented.
    def gen(
        self,
        n: Optional[int] = None,
        pending_observations: Optional[Dict[str, List[ObservationFeatures]]] = None,
        max_gen_draws_for_deduplication: int = MAX_GEN_DRAWS,
        arms_by_signature_for_deduplication: Optional[Dict[str, Arm]] = None,
        **model_gen_kwargs: Any,
    ) -> GeneratorRun:
        """This method generates candidates using `self._gen` and handles deduplication
        of generated candidates if `self.should_deduplicate=True`.

        NOTE: Models must have been fit prior to calling ``gen``.
        NOTE: Some underlying models may ignore the ``n`` argument and produce a
            model-determined number of arms. In that case this method will also output
            a generator run with number of arms that may differ from ``n``.

        Args:
            n: Optional integer representing how many arms should be in the generator
                run produced by this method. When this is ``None``, ``n`` will be
                determined by the ``ModelSpec`` that we are generating from.
            pending_observations: A map from metric name to pending
                observations for that metric, used by some models to avoid
                resuggesting points that are currently being evaluated.
            max_gen_draws_for_deduplication: Maximum number of attempts for generating
                new candidates without duplicates. If non-duplicate candidates are not
                generated with these attempts, a ``GenerationStrategyRepeatedPoints``
                exception will be raised.
            arms_by_signature_for_deduplication: A dictionary mapping arm signatures to
                the arms, to be used for deduplicating newly generated arms.
            model_gen_kwargs: Keyword arguments, passed through to ``ModelSpec.gen``;
                these override any pre-specified in ``ModelSpec.model_gen_kwargs``.

        Returns:
            A ``GeneratorRun`` containing the newly generated candidates.
        """
        should_generate_run = True
        generator_run = None
        n_gen_draws = 0
        # Keep generating until each of `generator_run.arms` is not a duplicate
        # of a previous arm, if `should_deduplicate is True`
        while should_generate_run:
            generator_run = self._gen(
                n=n,
                pending_observations=pending_observations,
                **model_gen_kwargs,
            )
            should_generate_run = (
                self.should_deduplicate
                and arms_by_signature_for_deduplication
                and any(
                    arm.signature in arms_by_signature_for_deduplication
                    for arm in generator_run.arms
                )
            )
            n_gen_draws += 1
            if should_generate_run:
                if n_gen_draws > max_gen_draws_for_deduplication:
                    raise GenerationStrategyRepeatedPoints(
                        MAX_GEN_DRAWS_EXCEEDED_MESSAGE
                    )
                else:
                    logger.info(
                        "The generator run produced duplicate arms. Re-running the "
                        "generation step in an attempt to deduplicate. Candidates "
                        f"produced in the last generator run: {generator_run.arms}."
                    )
        assert generator_run is not None, (
            "The GeneratorRun is None which is an unexpected state of this"
            " GenerationStrategy. This occurred on GenerationNode: {self.node_name}."
        )
        generator_run._generation_node_name = self.node_name
        return generator_run

    def _gen(
        self,
        n: Optional[int] = None,
        pending_observations: Optional[Dict[str, List[ObservationFeatures]]] = None,
        **model_gen_kwargs: Any,
    ) -> GeneratorRun:
        """Picks a fitted model, from which to generate candidates (via
        ``self._pick_fitted_model_to_gen_from``) and generates candidates
        from it. Uses the ``model_gen_kwargs`` set on the selected ``ModelSpec``
        alongside any kwargs passed in to this function (with local kwargs)
        taking precedent.

        Args:
            n: Optional integer representing how many arms should be in the generator
                run produced by this method. When this is ``None``, ``n`` will be
                determined by the ``ModelSpec`` that we are generating from.
            pending_observations: A map from metric name to pending
                observations for that metric, used by some models to avoid
                resuggesting points that are currently being evaluated.
            model_gen_kwargs: Keyword arguments, passed through to ``ModelSpec.gen``;
                these override any pre-specified in ``ModelSpec.model_gen_kwargs``.

        Returns:
            A ``GeneratorRun`` containing the newly generated candidates.
        """
        model_spec = self.model_spec_to_gen_from
        if n is None and model_spec.model_gen_kwargs:
            # If `n` is not specified, ensure that the `None` value does not
            # override the one set in `model_spec.model_gen_kwargs`.
            n = model_spec.model_gen_kwargs.get("n", None)
        return model_spec.gen(
            n=n,
            # For `pending_observations`, prefer the input to this function, as
            # `pending_observations` are dynamic throughout the experiment and thus
            # unlikely to be specified in `model_spec.model_gen_kwargs`.
            pending_observations=pending_observations,
            **model_gen_kwargs,
        )

    # ------------------------- Model selection logic helpers. -------------------------

    def _pick_fitted_model_to_gen_from(self) -> ModelSpec:
        """Select one model to generate from among the fitted models on this
        generation node.

        NOTE: In base ``GenerationNode`` class, this method does the following:
          1. if this ``GenerationNode`` has an associated ``BestModelSelector``,
             use it to select one model to generate from among the fitted models
             on this generation node.
          2. otherwise, ensure that this ``GenerationNode`` only contains one
             `ModelSpec` and select it.
        """
        if self.best_model_selector is None:
            if len(self.model_specs) != 1:
                raise NotImplementedError(CANNOT_SELECT_ONE_MODEL_MSG)
            return self.model_specs[0]

        best_model_index = not_none(self.best_model_selector).best_model(
            model_specs=self.model_specs,
        )
        return self.model_specs[best_model_index]

    # ------------------------- Trial logic helpers. -------------------------
    @property
    def trials_from_node(self) -> Set[int]:
        """Returns a set mapping a GenerationNode to the trials it generated.

        Returns:
            Set[int]: A set containing all the trials indices generated by this node.
        """
        trials_from_node = set()
        for _idx, trial in self.experiment.trials.items():
            for gr in trial.generator_runs:
                if (
                    gr._generation_node_name is not None
                    and gr._generation_node_name == self.node_name
                ):
                    trials_from_node.add(trial.index)
        return trials_from_node

    @property
    def node_that_generated_last_gr(self) -> Optional[str]:
        """Returns the name of the node that generated the last generator run.

        Returns:
            str: The name of the node that generated the last generator run.
        """
        return (
            self.generation_strategy.last_generator_run._generation_node_name
            if self.generation_strategy.last_generator_run
            else None
        )

    @property
    def transition_edges(self) -> Dict[str, List[TransitionCriterion]]:
        """Returns a dictionary mapping the next ``GenerationNode`` to the
        TransitionCriteria that define the transition that that node.

        Ex: if the transition from the current node to node x is defined by MaxTrials
        and MinTrials criterion then the return would be {'x': [MaxTrials, MinTrials]}.

        Returns:
            Dict[str, List[TransitionCriterion]]: A dictionary mapping the next
            ``GenerationNode`` to the ``TransitionCriterion`` that are associated
            with it.
        """
        if self.transition_criteria is None:
            return {}

        tc_edges = defaultdict(list)
        for tc in self.transition_criteria:
            tc_edges[tc.transition_to].append(tc)
        return tc_edges

    def should_transition_to_next_node(
        self, raise_data_required_error: bool = True
    ) -> Tuple[bool, Optional[str]]:
        """Checks whether we should transition to the next node based on this node's
        TransitionCriterion.

        Important: This method relies on the ``transition_criterion`` of this node to
        be listed in order of importance. Ex: a fallback transition should come after
        the primary transition in the transition criterion list.

        Args:
            raise_data_required_error: Whether to raise ``DataRequiredError`` in the
                case detailed above. Not raising the error is useful if just looking to
                check how many generator runs (to be made into trials) can be produced,
                but not actually producing them yet.
        Returns:
            Tuple[bool, Optional[str]]: Whether we should transition to the next node
                and the name of the next node.
        """
        # if no transition criteria are defined, this node can generate unlimited trials
        if len(self.transition_criteria) == 0:
            return False, None

        # for each edge in node DAG, check if the transition criterion are met, if so
        # transition to the next node defined by that edge.
        for next_node, all_tc in self.transition_edges.items():
            transition_blocking = [tc for tc in all_tc if tc.block_transition_if_unmet]
            gs_lgr = self.generation_strategy.last_generator_run
            transition_blocking_met = all(
                tc.is_met(
                    experiment=self.experiment,
                    trials_from_node=self.trials_from_node,
                    curr_node_name=self.node_name,
                    # TODO @mgarrard: should we instead pass a backpointer to gs/node
                    node_that_generated_last_gr=(
                        gs_lgr._generation_node_name if gs_lgr is not None else None
                    ),
                )
                for tc in transition_blocking
            )

            # Raise any necessary generation errors: for any met criterion,
            # call its `block_continued_generation_error` method if not all
            # transition-blocking criteria are met. The method might not raise an
            # error, depending on its implementation on given criterion, so the error
            # from the first met one that does block continued generation, will raise.
            # TODO: @mgarrard see if we can replace MaxGenerationParallelism with a
            # transition to self and rework this error block.
            if not transition_blocking_met:
                for tc in all_tc:
                    if (
                        tc.is_met(
                            self.experiment, trials_from_node=self.trials_from_node
                        )
                        and raise_data_required_error
                    ):
                        tc.block_continued_generation_error(
                            node_name=self.node_name,
                            model_name=self.model_to_gen_from_name,
                            experiment=self.experiment,
                            trials_from_node=self.trials_from_node,
                        )
            if len(transition_blocking) > 0 and transition_blocking_met:
                return True, next_node

        return False, None

    def generator_run_limit(self, raise_generation_errors: bool = False) -> int:
        """How many generator runs can this generation strategy generate right now,
        assuming each one of them becomes its own trial. Only considers
        `transition_criteria` that are TrialBasedCriterion.

        Returns:
            The number of generator runs that can currently be produced, with -1
            meaning unlimited generator runs.
        """
        # TODO: @mgarrard Should we consider returning `None` if there is no limit?
        trial_based_gen_blocking_criteria = [
            criterion
            for criterion in self.transition_criteria
            if criterion.block_gen_if_met and isinstance(criterion, TrialBasedCriterion)
        ]
        gen_blocking_criterion_delta_from_threshold = [
            criterion.num_till_threshold(
                experiment=self.experiment, trials_from_node=self.trials_from_node
            )
            for criterion in trial_based_gen_blocking_criteria
        ]

        # Raise any necessary generation errors: for any met criterion,
        # call its `block_continued_generation_error` method The method might not
        # raise an error, depending on its implementation on given criterion, so the
        # error from the first met one that does block continued generation, will be
        # raised.
        if raise_generation_errors:
            for criterion in trial_based_gen_blocking_criteria:
                # TODO[mgarrard]: Raise a group of all the errors, from each gen-
                # blocking transition criterion.
                if criterion.is_met(
                    self.experiment, trials_from_node=self.trials_from_node
                ):
                    criterion.block_continued_generation_error(
                        node_name=self.node_name,
                        model_name=self.model_to_gen_from_name,
                        experiment=self.experiment,
                        trials_from_node=self.trials_from_node,
                    )
        if len(gen_blocking_criterion_delta_from_threshold) == 0:
            return -1
        return min(gen_blocking_criterion_delta_from_threshold)

    def __repr__(self) -> str:
        "String representation of this GenerationNode"
        # add model specs
        str_rep = f"{self.__class__.__name__}(model_specs="
        model_spec_str = str(self.model_specs).replace("\n", " ").replace("\t", "")
        str_rep += model_spec_str

        str_rep += f", node_name={self.node_name}"
        str_rep += f", transition_criteria={str(self.transition_criteria)}"

        return f"{str_rep})"


@dataclass
class GenerationStep(GenerationNode, SortableBase):
    """One step in the generation strategy, corresponds to a single model.
    Describes the model, how many trials will be generated with this model, what
    minimum number of observations is required to proceed to the next model, etc.

    NOTE: Model can be specified either from the model registry
    (`ax.modelbridge.registry.Models` or using a callable model constructor. Only
    models from the registry can be saved, and thus optimization can only be
    resumed if interrupted when using models from the registry.

    Args:
        model: A member of `Models` enum or a callable returning an instance of
            `ModelBridge` with an instantiated underlying `Model`. Refer to
            `ax/modelbridge/factory.py` for examples of such callables.
        num_trials: How many trials to generate with the model from this step.
            If set to -1, trials will continue to be generated from this model
            as long as `generation_strategy.gen` is called (available only for
            the last of the generation steps).
        min_trials_observed: How many trials must be completed before the
            generation strategy can proceed to the next step. Defaults to 0.
            If `num_trials` of a given step have been generated but `min_trials_
            observed` have not been completed, a call to `generation_strategy.gen`
            will fail with a `DataRequiredError`.
        max_parallelism: How many trials generated in the course of this step are
            allowed to be run (i.e. have `trial.status` of `RUNNING`) simultaneously.
            If `max_parallelism` trials from this step are already running, a call
            to `generation_strategy.gen` will fail with a `MaxParallelismReached
            Exception`, indicating that more trials need to be completed before
            generating and running next trials.
        use_update: DEPRECATED.
        enforce_num_trials: Whether to enforce that only `num_trials` are generated
            from the given step. If False and `num_trials` have been generated, but
            `min_trials_observed` have not been completed, `generation_strategy.gen`
            will continue generating trials from the current step, exceeding `num_
            trials` for it. Allows to avoid `DataRequiredError`, but delays
            proceeding to next generation step.
        model_kwargs: Dictionary of kwargs to pass into the model constructor on
            instantiation. E.g. if `model` is `Models.SOBOL`, kwargs will be applied
            as `Models.SOBOL(**model_kwargs)`; if `model` is `get_sobol`, `get_sobol(
            **model_kwargs)`. NOTE: if generation strategy is interrupted and
            resumed from a stored snapshot and its last used model has state saved on
            its generator runs, `model_kwargs` is updated with the state dict of the
            model, retrieved from the last generator run of this generation strategy.
        model_gen_kwargs: Each call to `generation_strategy.gen` performs a call to the
            step's model's `gen` under the hood; `model_gen_kwargs` will be passed to
            the model's `gen` like so: `model.gen(**model_gen_kwargs)`.
        completion_criteria: List of TransitionCriterion. All `is_met` must evaluate
            True for the GenerationStrategy to move on to the next Step
        index: Index of this generation step, for use internally in `Generation
            Strategy`. Do not assign as it will be reassigned when instantiating
            `GenerationStrategy` with a list of its steps.
        should_deduplicate: Whether to deduplicate the parameters of proposed arms
            against those of previous arms via rejection sampling. If this is True,
            the generation strategy will discard generator runs produced from the
            generation step that has `should_deduplicate=True` if they contain arms
            already present on the experiment and replace them with new generator runs.
            If no generator run with entirely unique arms could be produced in 5
            attempts, a `GenerationStrategyRepeatedPoints` error will be raised, as we
            assume that the optimization converged when the model can no longer suggest
            unique arms.

    Note for developers: by "model" here we really mean an Ax ModelBridge object, which
    contains an Ax Model under the hood. We call it "model" here to simplify and focus
    on explaining the logic of GenerationStep and GenerationStrategy.
    """

    # Required options:
    model: Union[ModelRegistryBase, Callable[..., ModelBridge]]
    num_trials: int

    # Optional model specifications:
    # Kwargs to pass into the Models constructor (or factory function).
    model_kwargs: Dict[str, Any] = field(default_factory=dict)
    # Kwargs to pass into the Model's `.gen` function.
    model_gen_kwargs: Dict[str, Any] = field(default_factory=dict)

    # Optional specifications for use in generation strategy:
    completion_criteria: Sequence[TransitionCriterion] = field(default_factory=list)
    min_trials_observed: int = 0
    max_parallelism: Optional[int] = None
    use_update: bool = False
    enforce_num_trials: bool = True
    # Whether the generation strategy should deduplicate the suggested arms against
    # the arms already present on the experiment. If this is `True`
    # on a given generation step, during that step the generation
    # strategy will discard a generator run that contains an arm
    # already present on the experiment and produce a new generator
    # run instead before returning it from `gen` or `_gen_multiple`.
    should_deduplicate: bool = False
    index: int = -1  # Index of this step, set internally.

    # Optional model name. Defaults to `model_spec.model_key`.
    model_name: str = field(default_factory=str)

    def __post_init__(self) -> None:
        if self.use_update:
            raise DeprecationWarning("`GenerationStep.use_update` is deprecated.")
        if (
            self.enforce_num_trials
            and (self.num_trials >= 0)
            and (self.min_trials_observed > self.num_trials)
        ):
            raise UserInputError(
                "`GenerationStep` received `min_trials_observed > num_trials` "
                f"(`min_trials_observed = {self.min_trials_observed}`, `num_trials = "
                f"{self.num_trials}`), making completion of this step impossible. "
                "Please alter inputs so that `min_trials_observed <= num_trials`."
            )
        # For backwards compatibility with None / Optional input.
        self.model_kwargs = self.model_kwargs if self.model_kwargs is not None else {}
        self.model_gen_kwargs = (
            self.model_gen_kwargs if self.model_gen_kwargs is not None else {}
        )
        if not isinstance(self.model, ModelRegistryBase):
            if not callable(self.model):
                raise UserInputError(
                    "`model` in generation step must be either a `ModelRegistryBase` "
                    "enum subclass entry or a callable factory function returning a "
                    "model bridge instance."
                )
            model_spec = FactoryFunctionModelSpec(
                factory_function=self.model,
                model_kwargs=self.model_kwargs,
                model_gen_kwargs=self.model_gen_kwargs,
            )
        else:
            model_spec = ModelSpec(
                model_enum=self.model,
                model_kwargs=self.model_kwargs,
                model_gen_kwargs=self.model_gen_kwargs,
            )
        if self.model_name == "":
            try:
                self.model_name = model_spec.model_key
            except TypeError:
                # Factory functions may not always have a model key defined.
                self.model_name = f"Unknown {model_spec.__class__.__name__}"

        # Create transition criteria for this step. MaximumTrialsInStatus can be used
        # to ensure that requirements related to num_trials and unlimited trials
        # are met. MinimumTrialsInStatus can be used enforce the min_trials_observed
        # requirement, and override MaxTrials if enforce flag is set to true. We set
        # `transition_to` is set in `GenerationStrategy` constructor,
        # because only then is the order of the generation steps actually known.
        transition_criteria = []
        if self.num_trials != -1:
            transition_criteria.append(
                MaxTrials(
                    threshold=self.num_trials,
                    not_in_statuses=[TrialStatus.FAILED, TrialStatus.ABANDONED],
                    block_gen_if_met=self.enforce_num_trials,
                    block_transition_if_unmet=True,
                )
            )

        if self.min_trials_observed > 0:
            transition_criteria.append(
                MinTrials(
                    only_in_statuses=[
                        TrialStatus.COMPLETED,
                        TrialStatus.EARLY_STOPPED,
                    ],
                    threshold=self.min_trials_observed,
                    block_gen_if_met=False,
                    block_transition_if_unmet=True,
                )
            )
        if self.max_parallelism is not None:
            transition_criteria.append(
                MaxGenerationParallelism(
                    threshold=self.max_parallelism,
                    only_in_statuses=[TrialStatus.RUNNING],
                    block_gen_if_met=True,
                    block_transition_if_unmet=False,
                    transition_to=None,
                )
            )

        transition_criteria += self.completion_criteria
        super().__init__(
            node_name=f"GenerationStep_{str(self.index)}",
            model_specs=[model_spec],
            should_deduplicate=self.should_deduplicate,
            transition_criteria=transition_criteria,
        )

    @property
    def model_spec(self) -> ModelSpec:
        """Returns the first model_spec from the model_specs attribute."""
        return self.model_specs[0]

    @property
    def _unique_id(self) -> str:
        """Returns the unique ID of this generation step, which is the index."""
        return str(self.index)

    def gen(
        self,
        n: Optional[int] = None,
        pending_observations: Optional[Dict[str, List[ObservationFeatures]]] = None,
        max_gen_draws_for_deduplication: int = MAX_GEN_DRAWS,
        arms_by_signature_for_deduplication: Optional[Dict[str, Arm]] = None,
        **model_gen_kwargs: Any,
    ) -> GeneratorRun:
        gr = super().gen(
            n=n,
            pending_observations=pending_observations,
            max_gen_draws_for_deduplication=max_gen_draws_for_deduplication,
            arms_by_signature_for_deduplication=arms_by_signature_for_deduplication,
            **model_gen_kwargs,
        )
        gr._generation_step_index = self.index
        return gr

    def __eq__(self, other: Base) -> bool:
        # We need to override `__eq__` to make sure we inherit the one from
        # the base class and not the one from dataclasses library, since we
        # want to be comparing equality of generation steps in the same way
        # as we compare equality of other Ax objects (and we want all the
        # same special-casing to apply).
        return SortableBase.__eq__(self, other=other)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import Any, Dict, Optional
from warnings import warn

from ax.modelbridge.cross_validation import (
    get_fit_and_std_quality_and_generalization_dict,
)

from ax.service.scheduler import get_fitted_model_bridge, Scheduler
from ax.telemetry.common import _get_max_transformed_dimensionality

from ax.telemetry.experiment import ExperimentCompletedRecord, ExperimentCreatedRecord
from ax.telemetry.generation_strategy import GenerationStrategyCreatedRecord


@dataclass(frozen=True)
class SchedulerCreatedRecord:
    """
    Record of the Scheduler creation event. This can be used for telemetry in settings
    where many Schedulers are being created either manually or programatically. In
    order to facilitate easy serialization only include simple types: numbers, strings,
    bools, and None.
    """

    experiment_created_record: ExperimentCreatedRecord
    generation_strategy_created_record: GenerationStrategyCreatedRecord

    # SchedulerOptions info
    scheduler_total_trials: Optional[int]
    scheduler_max_pending_trials: int
    arms_per_trial: int
    early_stopping_strategy_cls: Optional[str]
    global_stopping_strategy_cls: Optional[str]

    # Dimensionality of transformed SearchSpace can often be much higher due to one-hot
    # encoding of unordered ChoiceParameters
    transformed_dimensionality: Optional[int]

    @classmethod
    def from_scheduler(cls, scheduler: Scheduler) -> SchedulerCreatedRecord:
        return cls(
            experiment_created_record=ExperimentCreatedRecord.from_experiment(
                experiment=scheduler.experiment
            ),
            generation_strategy_created_record=(
                GenerationStrategyCreatedRecord.from_generation_strategy(
                    generation_strategy=scheduler.standard_generation_strategy
                )
            ),
            scheduler_total_trials=scheduler.options.total_trials,
            scheduler_max_pending_trials=scheduler.options.max_pending_trials,
            # If batch_size is None then we are using single-Arm trials
            arms_per_trial=scheduler.options.batch_size or 1,
            early_stopping_strategy_cls=(
                None
                if scheduler.options.early_stopping_strategy is None
                else scheduler.options.early_stopping_strategy.__class__.__name__
            ),
            global_stopping_strategy_cls=(
                None
                if scheduler.options.global_stopping_strategy is None
                else scheduler.options.global_stopping_strategy.__class__.__name__
            ),
            transformed_dimensionality=_get_max_transformed_dimensionality(
                search_space=scheduler.experiment.search_space,
                generation_strategy=scheduler.standard_generation_strategy,
            ),
        )

    def flatten(self) -> Dict[str, Any]:
        """
        Flatten into an appropriate format for logging to a tabular database.
        """

        self_dict = asdict(self)
        experiment_created_record_dict = self_dict.pop("experiment_created_record")
        generation_strategy_created_record_dict = self_dict.pop(
            "generation_strategy_created_record"
        )

        return {
            **self_dict,
            **experiment_created_record_dict,
            **generation_strategy_created_record_dict,
        }


@dataclass(frozen=True)
class SchedulerCompletedRecord:
    """
    Record of the Scheduler completion event. This will have information only available
    after the optimization has completed.
    """

    experiment_completed_record: ExperimentCompletedRecord

    best_point_quality: float
    model_fit_quality: Optional[float]
    model_std_quality: Optional[float]
    model_fit_generalization: Optional[float]
    model_std_generalization: Optional[float]

    improvement_over_baseline: float

    num_metric_fetch_e_encountered: int
    num_trials_bad_due_to_err: int

    @classmethod
    def from_scheduler(cls, scheduler: Scheduler) -> SchedulerCompletedRecord:
        try:
            model_bridge = get_fitted_model_bridge(scheduler)
            quality_and_generalizations_dict = (
                get_fit_and_std_quality_and_generalization_dict(
                    fitted_model_bridge=model_bridge,
                )
            )
        except Exception as e:
            warn("Encountered exception in computing model fit quality: " + str(e))
            quality_and_generalizations_dict = {
                "model_fit_quality": None,
                "model_std_quality": None,
                "model_fit_generalization": None,
                "model_std_generalization": None,
            }

        try:
            improvement_over_baseline = scheduler.get_improvement_over_baseline()
        except Exception as e:
            warn(
                "Encountered exception in computing improvement over baseline: "
                + str(e)
            )
            improvement_over_baseline = float("nan")

        return cls(
            experiment_completed_record=ExperimentCompletedRecord.from_experiment(
                experiment=scheduler.experiment
            ),
            best_point_quality=float("nan"),  # TODO[T147907632]
            improvement_over_baseline=improvement_over_baseline,
            num_metric_fetch_e_encountered=scheduler._num_metric_fetch_e_encountered,
            num_trials_bad_due_to_err=scheduler._num_trials_bad_due_to_err,
            **quality_and_generalizations_dict,
        )

    def flatten(self) -> Dict[str, Any]:
        """
        Flatten into an appropriate format for logging to a tabular database.
        """

        self_dict = asdict(self)
        experiment_completed_record_dict = self_dict.pop("experiment_completed_record")

        return {
            **self_dict,
            **experiment_completed_record_dict,
        }



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import math
from typing import Tuple

import pandas as pd
import plotly.graph_objs as go
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.modelbridge.factory import get_empirical_bayes_thompson, get_thompson
from ax.plot.base import AxPlotConfig, AxPlotTypes, PlotMetric, Z
from ax.plot.helper import get_plot_data
from ax.plot.scatter import _error_scatter_data
from pandas.core.frame import DataFrame


COLOR_SCALE = ["#ff3333", "#ff6666", "#ffffff", "#99ff99", "#33ff33"]


def get_color(x: float, ci: float, rel: bool, reverse: bool) -> str:
    """Determine the color of the table cell."""
    if not rel:
        # Color coding is meant to be relative to the status quo,
        # and thus doesn't make sense if rel = False
        return "#ffffff"

    r = min(math.floor(abs(x) / ci), 2) if ci > 0 else 2
    index = int(2 + r * math.copysign(1, x))
    color_scale = list(COLOR_SCALE)
    if reverse:
        color_scale = list(reversed(color_scale))
    return color_scale[index]


def table_view_plot(
    experiment: Experiment,
    data: Data,
    use_empirical_bayes: bool = True,
    only_data_frame: bool = False,
    arm_noun: str = "arm",
) -> Tuple[DataFrame]:
    """Table of means and confidence intervals.

    Table is of the form:

    +-------+------------+-----------+
    |  arm  |  metric_1  |  metric_2 |
    +=======+============+===========+
    |  0_0  | mean +- CI |    ...    |
    +-------+------------+-----------+
    |  0_1  |    ...     |    ...    |
    +-------+------------+-----------+

    """
    model_func = get_empirical_bayes_thompson if use_empirical_bayes else get_thompson
    model = model_func(experiment=experiment, data=data)

    # We don't want to include metrics from a collection,
    # or the chart will be too big to read easily.
    # Example:
    # experiment.metrics = {
    #   'regular_metric': Metric(),
    #   'collection_metric: CollectionMetric()', # collection_metric =[metric1, metric2]
    # }
    # model.metric_names = [regular_metric, metric1, metric2] # "exploded" out
    # We want to filter model.metric_names and get rid of metric1, metric2
    metric_names = [
        metric_name
        for metric_name in model.metric_names
        if metric_name in experiment.metrics
    ]

    metric_name_to_lower_is_better = {
        metric_name: experiment.metrics[metric_name].lower_is_better
        for metric_name in metric_names
    }

    plot_data, _, _ = get_plot_data(
        model=model,
        generator_runs_dict={},
        # pyre-fixme[6]: Expected `Optional[typing.Set[str]]` for 3rd param but got
        #  `List[str]`.
        metric_names=metric_names,
    )

    if plot_data.status_quo_name:
        status_quo_arm = plot_data.in_sample.get(plot_data.status_quo_name)
        rel = True
    else:
        status_quo_arm = None
        rel = False

    records = []
    colors = []
    records_with_mean = []
    records_with_ci = []
    for metric_name in metric_names:
        arm_names, _, ys, ys_se = _error_scatter_data(
            arms=list(plot_data.in_sample.values()),
            y_axis_var=PlotMetric(metric_name, pred=True, rel=rel),
            x_axis_var=None,
            status_quo_arm=status_quo_arm,
        )

        results_by_arm = list(zip(arm_names, ys, ys_se))
        colors.append(
            [
                get_color(
                    x=y,
                    ci=Z * y_se,
                    rel=rel,
                    # pyre-fixme[6]: Expected `bool` for 4th param but got
                    #  `Optional[bool]`.
                    reverse=metric_name_to_lower_is_better[metric_name],
                )
                for (_, y, y_se) in results_by_arm
            ]
        )
        records.append(
            [
                "{:.3f} &plusmn; {:.3f}".format(y, Z * y_se)
                for (_, y, y_se) in results_by_arm
            ]
        )
        records_with_mean.append({arm_name: y for (arm_name, y, _) in results_by_arm})
        records_with_ci.append(
            {arm_name: Z * y_se for (arm_name, _, y_se) in results_by_arm}
        )

    if only_data_frame:
        return tuple(
            pd.DataFrame.from_records(records, index=metric_names)
            for records in [records_with_mean, records_with_ci]
        )

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def transpose(m):
        return [[m[j][i] for j in range(len(m))] for i in range(len(m[0]))]

    records = [[name.replace(":", " : ") for name in metric_names]] + transpose(records)
    colors = [["#ffffff"] * len(metric_names)] + transpose(colors)
    # pyre-fixme[61]: `arm_names` may not be initialized here.
    header = [f"<b>{x}</b>" for x in [f"{arm_noun}s"] + arm_names]
    # pyre-fixme[61]: `arm_names` may not be initialized here.
    column_widths = [300] + [150] * len(arm_names)

    trace = go.Table(
        header={"values": header, "align": ["left"]},
        cells={"values": records, "align": ["left"], "fill": {"color": colors}},
        columnwidth=column_widths,
    )
    layout = go.Layout(
        width=sum(column_widths),
        margin=go.layout.Margin(l=0, r=20, b=20, t=20, pad=4),  # noqa E741
    )
    fig = go.Figure(data=[trace], layout=layout)
    # pyre-fixme[7]: Expected `Tuple[DataFrame]` but got `AxPlotConfig`.
    return AxPlotConfig(data=fig, plot_type=AxPlotTypes.GENERIC)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import os
import shutil
import tempfile
from typing import List

from ax.core import (
    BatchTrial,
    Experiment,
    Objective,
    OptimizationConfig,
    Parameter,
    ParameterType,
    RangeParameter,
    SearchSpace,
)
from ax.metrics.torchx import TorchXMetric
from ax.modelbridge.dispatch_utils import choose_generation_strategy
from ax.runners.torchx import TorchXRunner
from ax.service.scheduler import FailureRateExceededError, Scheduler, SchedulerOptions
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from torchx.components import utils


class TorchXRunnerTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.test_dir = tempfile.mkdtemp("torchx_runtime_hpo_ax_test")

        self.old_cwd = os.getcwd()
        os.chdir(os.path.dirname(os.path.dirname(__file__)))

        self._parameters: List[Parameter] = [
            RangeParameter(
                name="x1",
                lower=-10.0,
                upper=10.0,
                parameter_type=ParameterType.FLOAT,
            ),
            RangeParameter(
                name="x2",
                lower=-10.0,
                upper=10.0,
                parameter_type=ParameterType.FLOAT,
            ),
        ]

        self._minimize = True
        self._objective = Objective(
            metric=TorchXMetric(
                name="booth_eval",
            ),
            minimize=self._minimize,
        )

        self._runner = TorchXRunner(
            tracker_base=self.test_dir,
            component=utils.booth,
            scheduler="local_cwd",
            cfg={"prepend_cwd": True},
        )

    def tearDown(self) -> None:
        shutil.rmtree(self.test_dir)
        os.chdir(self.old_cwd)

    def test_run_experiment_locally(self) -> None:
        """Runs optimization over n rounds of k sequential trials."""

        experiment = Experiment(
            name="torchx_booth_sequential_demo",
            search_space=SearchSpace(parameters=self._parameters),
            optimization_config=OptimizationConfig(objective=self._objective),
            runner=self._runner,
            is_test=True,
            properties={Keys.IMMUTABLE_SEARCH_SPACE_AND_OPT_CONF: True},
        )

        scheduler = Scheduler(
            experiment=experiment,
            generation_strategy=(
                choose_generation_strategy(
                    search_space=experiment.search_space,
                )
            ),
            options=SchedulerOptions(),
        )

        try:
            for _ in range(3):
                scheduler.run_n_trials(max_trials=2)

            # TorchXMetric always returns trial index; hence the best experiment
            # for min objective will be the params for trial 0.
            scheduler.report_results()
        except FailureRateExceededError:
            pass  # TODO(ehotaj): Figure out why this test fails in OSS.
        # Nothing to assert, just make sure experiment runs.

    def test_stop_trials(self) -> None:
        experiment = Experiment(
            name="torchx_booth_sequential_demo",
            search_space=SearchSpace(parameters=self._parameters),
            optimization_config=OptimizationConfig(objective=self._objective),
            runner=self._runner,
            is_test=True,
            properties={Keys.IMMUTABLE_SEARCH_SPACE_AND_OPT_CONF: True},
        )
        scheduler = Scheduler(
            experiment=experiment,
            generation_strategy=(
                choose_generation_strategy(
                    search_space=experiment.search_space,
                )
            ),
            options=SchedulerOptions(),
        )
        scheduler.run(max_new_trials=3)
        trial = scheduler.running_trials[0]
        reason = self._runner.stop(trial, reason="some_reason")
        self.assertEqual(reason, {"reason": "some_reason"})

    def test_run_experiment_locally_in_batches(self) -> None:
        """Runs optimization over k x n rounds of k parallel trials.

        This asks Ax to run up to max_parallelism_cap trials in parallel by
        submitting them to the scheduler at the same time.

        NOTE:
            * setting max_parallelism_cap in generation_strategy
            * setting run_trials_in_batches in scheduler options
            * setting total_trials = parallelism * rounds

        """
        parallelism = 2
        rounds = 3

        experiment = Experiment(
            name="torchx_booth_parallel_demo",
            search_space=SearchSpace(parameters=self._parameters),
            optimization_config=OptimizationConfig(objective=self._objective),
            runner=self._runner,
            is_test=True,
            properties={Keys.IMMUTABLE_SEARCH_SPACE_AND_OPT_CONF: True},
        )

        scheduler = Scheduler(
            experiment=experiment,
            generation_strategy=(
                choose_generation_strategy(
                    search_space=experiment.search_space,
                    max_parallelism_cap=parallelism,
                )
            ),
            options=SchedulerOptions(
                run_trials_in_batches=True, total_trials=(parallelism * rounds)
            ),
        )

        try:
            scheduler.run_all_trials()

            # TorchXMetric always returns trial index; hence the best experiment
            # for min objective will be the params for trial 0.
            scheduler.report_results()
        except FailureRateExceededError:
            pass  # TODO(ehotaj): Figure out why this test fails in OSS.
        # Nothing to assert, just make sure experiment runs.

    def test_runner_no_batch_trials(self) -> None:
        experiment = Experiment(
            name="runner_test",
            search_space=SearchSpace(parameters=self._parameters),
            optimization_config=OptimizationConfig(objective=self._objective),
            runner=self._runner,
            is_test=True,
        )

        with self.assertRaises(ValueError):
            self._runner.run(trial=BatchTrial(experiment))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.core import (
    Arm,
    BatchTrial,
    ChoiceParameter,
    ComparisonOp,
    Data,
    Experiment,
    FixedParameter,
    GeneratorRun,
    Metric,
    MultiObjective,
    MultiObjectiveOptimizationConfig,
    Objective,
    ObjectiveThreshold,
    OptimizationConfig,
    OrderConstraint,
    OutcomeConstraint,
    Parameter,
    ParameterConstraint,
    ParameterType,
    RangeParameter,
    Runner,
    SearchSpace,
    SumConstraint,
    Trial,
)
from ax.modelbridge import Models
from ax.service import OptimizationLoop, optimize
from ax.storage import json_load, json_save

try:
    pass
except Exception:  # pragma: no cover
    __version__ = "Unknown"


__all__ = [
    "Arm",
    "BatchTrial",
    "ChoiceParameter",
    "ComparisonOp",
    "Data",
    "Experiment",
    "FixedParameter",
    "GeneratorRun",
    "Metric",
    "Models",
    "MultiObjective",
    "MultiObjectiveOptimizationConfig",
    "Objective",
    "ObjectiveThreshold",
    "OptimizationConfig",
    "OptimizationLoop",
    "OrderConstraint",
    "OutcomeConstraint",
    "Parameter",
    "ParameterConstraint",
    "ParameterType",
    "RangeParameter",
    "Runner",
    "SearchSpace",
    "SumConstraint",
    "Trial",
    "optimize",
    "json_save",
    "json_load",
]
